#include<stdio.h>
#include<math.h>
#include<stddef.h>
#include<stdlib.h>
#include<string.h>
#define TWOPI 6.28318531
#define FOPI 12.566371
extern float rth;
extern float rknyq,rkinv,powh,smooth,omehs;
extern float omep,omeplam,omepb,hubble,ssize;
extern char	dark[4];
float power(float rk){
	/*
      Calculates sqrt(Power spectrum) with gaussian smoothing and with
      biquadratic cutoff at Nyquist frequency
      Power spectrum consists of primordial scale-free power law and 
      transfer function
      Primordial scale-free power spectrum : power = sqrt(P(k))
      rk = 2*pi*(wave number) & rkinv = 1./(2*pi*knyq)
      smooth = (smoothing length in pixels /ng) /sqrt(2)
	*/
	float Power,qq,damp;
	if(rk > rknyq || rk == 0.) {
		Power = 0.;
		return Power;
	}
	else {
		/*
		 * power=rk**powh*(1.-(rk*rkinv)**4.)*exp(-(rk*smooth)**2.)
		 * */
		Power = pow(rk,powh);
	}
	/*
      Transfer function (BBKS 1986, p60) 
      omehs = 1/(omega*h*v[100])
	*/
	qq = rk*omehs;
	/*
      Cold dark matter, adiabatic fluctuations with 3 flavors of 
      relativistic neutrinos plus photons and negligible baryon
	  */
	if(strcmp(dark,"cdm")==0) {
		Power = Power*log(1.+2.34*qq)/(2.34*qq)*
			pow(1+3.89*qq+pow(16.1*qq,2)+pow(5.46*qq,3)+pow(6.71*qq,4),-0.25);
	}
	else if(strcmp(dark,"hdm")==0){
		/*
      	  Hot dark matter, adiabatic fluctuations with 2 species of
	      relativiatic neutrinos
		*/
		damp=2.6*qq;
		Power = Power*exp(-0.16*damp-0.5*damp*damp)/(1+1.6*qq
				+pow(4.*qq,1.5)+pow(0.92*qq,2));
	}
	return Power;
}
/*
c---  Integrand for calculating (dM/M)**2 in top hat window
*/
float power1(float);
float powint(float xk){
	float Powint;
	float rk,xk1;
	/* xk = wave number */
	rk = TWOPI*xk;
	/* Power spectrum(without knyquist cutoff) times volumn factor */
	Powint = FOPI*xk*xk*pow(power1(rk),2);
	/* Fourier transform of top-hat window function */
	xk1   = rth*rk;
	if(xk1 > 0.02) {
		Powint = Powint*9*pow(((sin(xk1)-xk1*cos(xk1))/pow(xk1,3)),2);
	}
	else
		Powint = Powint*pow(1-xk1*xk1*0.1,2);
	return Powint;
}
float bpower1(float);
/* c---  Integrand for calculating (dM/M)**2 in top hat window */
float bpowint(float xk){
	/* xk = wave number */
	float rk,xk1;
	float Bpowint,Powint;
	rk = TWOPI*xk;
	/* Power spectrum(without knyquist cutoff) times volumn factor */
	Powint = FOPI*xk*xk*pow(bpower1(rk),2);
	/* Fourier transform of top-hat window function */
	xk1 = rth*rk;
	if(xk1 > 0.02) {
		Bpowint = Powint*9*pow((sin(xk1)-xk1*cos(xk1))/pow(xk1,3),2);
	}
	else {
		Bpowint = Powint*pow((1-xk1*xk1*0.1),2);
	}
	return Bpowint;
}
/* c---  Integrand for calculating integral(P(k)*k^2) in top hat window */
float bpow2int(float xk){
	float rk,Powint,xk1;
	float Bpow2int;

	/* xk = wave number */
	rk = TWOPI*xk;
	/* Power spectrum times volumn factor */
	Powint = FOPI*xk*xk*rk*rk*pow(bpower1(rk),2);
	/* Fourier transform of top-hat window function */
	xk1 = rth * rk;
	if(xk1 > 0.02){
		Bpow2int = Powint*9*pow((sin(xk1)-xk1*cos(xk1))/pow(xk1,3),2);
	}
	else 
		Bpow2int = Powint*pow((1-xk1*xk1*0.1),2);
	return Bpow2int;
}
/* c---  Integrand for calculating integral(P(k)*k^2) in top hat window */
float pow2int(float xk){
	float rk,Powint;
	float Pow2int,xk1;

	/* xk = wave number */
	rk = TWOPI*xk;
	/* Power spectrum times volumn factor */
	Powint = FOPI*xk*xk*rk*rk*pow(bpower1(rk),2);
	/* Fourier transform of top-hat window function */
	xk1 = rth * rk;
	if(xk1 > 0.02){
		Pow2int = Powint*9*pow((sin(xk1)-xk1*cos(xk1))/pow(xk1,3),2);
	}
	else 
		Pow2int = Powint*pow((1-xk1*xk1*0.1),2);
	return Pow2int;
}

float power1(float rk){
	float Power1,qq;
	float Power,damp;
	/*
      Calculates sqrt(Power spectrum) with gaussian smoothing and with
      biquadratic cutoff at Nyquist frequency
      Power spectrum consists of primordial scale-free power law and 
      transfer function
      Primordial scale-free power spectrum : power = sqrt(P(k))
      rk = 2*pi*(wave number) & rkinv = 1./(2*pi*knyq)
      smooth = (smoothing length in pixels /ng) /sqrt(2)
	 * */
	Power = pow(rk,powh)*exp(-pow(rk*smooth,2));
	/*
      Transfer function (BBKS 1986, p60) 
      omehs = 1/(omega*h*v[100])
	 * */
	qq = rk*omehs;
	/*
      Cold dark matter, adiabatic fluctuations with 3 flavors of 
      relativistic neutrinos plus photons and negligible baryon
	 */
	if(strcmp(dark,"cdm")==0) {
		if(qq < 0.002) {
			Power = Power*(1-1.17*qq+pow(1.351*qq,2))*
				pow(1+3.89*qq+pow(16.1*qq,2)+pow(5.46*qq,3)+
				 pow(6.71*qq,4),-0.25);
		}
		else 
			Power=Power*log(1+2.34*qq)/(2.34*qq)*
				pow(1+3.89*qq+pow(16.1*qq,2)+pow(5.46*qq,3)+pow(6.71*qq,4),
						-0.25);
	}
	/*
      Hot dark matter, adiabatic fluctuations with 2 species of
      relativiatic neutrinos
	 */
	else if(strcmp(dark,"hdm")==0) {
        damp=2.6*qq;
        Power=Power*exp(-0.16*damp-0.5*damp*damp)/
			(1+1.6*qq+pow(4*qq,1.5)+pow(0.92*qq,2));
	}
	Power1 = Power;
	return Power1;
}

float tnot(float q, float alphac, float betac){
	float e,c,res;
	e = 2.718281828;
	c = 14.2/alphac+386./(1+69.9*pow(q,1.08));
	res = log(e+1.8*betac*q)/(log(e+1.8*betac*q)+c*q*q);
	return res;
}

/*
c---  Calculates sqrt(Power spectrum) with gaussian smoothing and with
c---  cutoff at Nyquist frequency
c---  Power spectrum consists of primordial scale-free power law and
c---  transfer function
c---  Ref. Eisenstein and Hu, 1998, ApJ, 496,605
*/
float bpower1(float rk){
	static int in1=0;
    static float cmbt27,o,ob,oc,zeq,rkeq,b1,b2,zd,Rd,Req,sound,rksk,h;
	static float y,gy,alphab,a11,a22,alphac,b11,b22,betac,betab,betan;
	static float transc,soundt,rkstilde,rkrksk,transb,q;
	float qq,Power,rks,f;
	float Bpower;
	/*
      Important cosmological scales and epochs
      Tcmb = 2.728+-0.0004K (Fixen et al. 1996) adopted
      rkeq = scale of particle horizon at equality epoch in unit of [h Mpc^-1]
      zd = drag epoch (baryon released from the drag of photons)
      sound = sound horizon at drag epoch in [h^-1 Mpc]
      rksk = Silk damping scale in unit of [h Mpc^-1]
	 */
	h = hubble;
	if(in1 == 0){
		cmbt27 = 2.728/2.7;
		o = omep*h*h;
		ob = omepb*h*h;
		oc = (omep-omepb)*h*h;
        zeq = 2.50e4*o/pow(cmbt27,4);
        rkeq = 7.46e-2*omep*h/(cmbt27*cmbt27);
        b1  = 0.313/pow(o,0.419)*(1+0.607*pow(o,0.674));
        b2  = 0.238*pow(o,0.223);
        zd  = 1291*pow(o,0.251)/(1+0.659*pow(o,0.828))*(1+b1*pow(ob,b2));
        Rd  = 31.5*ob/pow(cmbt27,4)/(zd/1000.);
        Req = 31.5*ob/pow(cmbt27,4)/(zeq/1000.);
        sound=2./3./rkeq*sqrt(6./Req)
              *log((sqrt(1+Rd)+sqrt(Rd+Req))/(1+sqrt(Req)));
        rksk = 1.6*pow(ob,0.52)*pow(o,0.73)*(1+pow(10.4*o,-0.95))/h;
  
        y = (1+zeq)/(1+zd);
        gy = y*(-6*sqrt(1+y)+(2+3*y)*log((sqrt(1+y)+1)/(sqrt(1+y)-1)));
        alphab = 2.07*rkeq*sound/pow(1+Rd,3./4.)*gy;

        a11  = pow(46.9*o,0.670)*(1+pow(32.1*o,-0.532));
        a22  = pow(12.0*o,0.424)*(1+pow(45.0*o,-0.582));
        alphac = pow(a11,-ob/o)*pow(a22,(-ob/o*ob/o*ob/o));
        b11  = 0.944/(1+pow(458*o,-0.708));
        b22  = 1./pow(0.395*o,0.0266);
        betac = 1./(1+b11*(pow(oc/o,b22)-1));
        betab = 0.5+ob/o+(3-2*ob/o)*sqrt(pow(17.2*o,2)+1);
        betan = 8.41*pow(o,0.435);
        in1 = 1;
	}
	/*
      Primordial scale-free power spectrum : power = sqrt(P(k))
      rk = 2*pi*(wave number) & rkinv = 1./(2*pi*knyq)
      smooth = (smoothing length in pixels /ng) /sqrt(2)
	 * */
	if(rk > rknyq || rk == 0.) {
		Bpower = 0.;
		return Bpower;
	}
	else {
		Power = pow(rk,powh);
	}
	/*  Baryonic effects included */
    q   = rk*omehs*pow(cmbt27,2);
    rks = rk/ssize*sound;
    f   = 1./(1+pow(rks/5.4,4));
    transc = f*tnot(q,1.0,betac)+(1-f)*tnot(q,alphac,betac);
    soundt = sound/pow(1+pow(betan/rks,3),1./3.);
    rkstilde = rk/ssize*soundt;
    rkrksk = rk/ssize/rksk;
    transb = (tnot(q,1.0,1.0)/(1+pow(rks/5.2,2))
			+alphab/(1.+pow(betab/rks,3))*exp(-pow(rkrksk,1.4)))
		*sin(rkstilde)/rkstilde;
    Bpower = Power*fabs(ob*transb+oc*transc)/o;
	/*
      Transfer function (BBKS 1986, p60)
      omehs = 1/(omega*h*v[100])
	 */
	qq = rk*omehs;
	/*
      Cold dark matter, adiabatic fluctuations with 3 flavors of
      relativistic neutrinos plus photons and negligible baryon
          power=power*alog(1+2.34*qq)/(2.34*qq)*(1+3.89*qq
     &     +(16.1*qq)**2+(5.46*qq)**3+(6.71*qq)**4)**(-0.25)
	 */
	return Bpower;
}
      

#define MBIG 1000000000
#define MSEED 161803398
#define MZ 0
#define FAC (1.0/MBIG)

float ran3(long *idum)
{
	static int inext,inextp;
	static long ma[56];
	static int iff=0;
	long mj,mk;
	int i,ii,k;

	if (*idum < 0 || iff == 0) {
		iff=1;
		mj=MSEED-(*idum < 0 ? -*idum : *idum);
		mj %= MBIG;
		ma[55]=mj;
		mk=1;
		for (i=1;i<=54;i++) {
			ii=(21*i) % 55;
			ma[ii]=mk;
			mk=mj-mk;
			if (mk < MZ) mk += MBIG;
			mj=ma[ii];
		}
		for (k=1;k<=4;k++)
			for (i=1;i<=55;i++) {
				ma[i] -= ma[1+(i+30) % 55];
				if (ma[i] < MZ) ma[i] += MBIG;
			}
		inext=0;
		inextp=31;
		*idum=1;
	}
	if (++inext == 56) inext=1;
	if (++inextp == 56) inextp=1;
	mj=ma[inext]-ma[inextp];
	if (mj < MZ) mj += MBIG;
	ma[inext]=mj;
	return mj*FAC;
}
#undef MBIG
#undef MSEED
#undef MZ
#undef FAC
/* (C) Copr. 1986-92 Numerical Recipes Software 71.+I0>+. */
#include <math.h>

float gasdev(long *idum)
{
	float ran1(long *idum);
	static int iset=0;
	static float gset;
	float fac,rsq,v1,v2;

	if  (iset == 0) {
		do {
			v1=2.0*ran1(idum)-1.0;
			v2=2.0*ran1(idum)-1.0;
			rsq=v1*v1+v2*v2;
		} while (rsq >= 1.0 || rsq == 0.0);
		fac=sqrt(-2.0*log(rsq)/rsq);
		gset=v1*fac;
		iset=1;
		return v2*fac;
	} else {
		iset=0;
		return gset;
	}
}
/* (C) Copr. 1986-92 Numerical Recipes Software 71.+I0>+. */
#define FUNC(x) ((*func)(x))

float trapzd(float (*func)(float), float a, float b, int n)
{
	float x,tnm,sum,del;
	static float s;
	int it,j;

	if (n == 1) {
		return (s=0.5*(b-a)*(FUNC(a)+FUNC(b)));
	} else {
		for (it=1,j=1;j<n-1;j++) it <<= 1;
		tnm=it;
		del=(b-a)/tnm;
		x=a+0.5*del;
		for (sum=0.0,j=1;j<=it;j++,x+=del) sum += FUNC(x);
		s=0.5*(s+(b-a)*sum/tnm);
		return s;
	}
}
#undef FUNC
/* (C) Copr. 1986-92 Numerical Recipes Software 71.+I0>+. */
#include <math.h>
#define EPS 1.0e-6
#define JMAX 20

float qsimp(float (*func)(float), float a, float b)
{
	float trapzd(float (*func)(float), float a, float b, int n);
	void nrerror(char error_text[]);
	int j;
	float s,st,ost,os;

	ost = os = -1.0e30;
	for (j=1;j<=JMAX;j++) {
		st=trapzd(func,a,b,j);
		s=(4.0*st-ost)/3.0;
		if (fabs(s-os) < EPS*fabs(os)) return s;
		os=s;
		ost=st;
	}
	nrerror("Too many steps in routine qsimp");
	return 0.0;
}
#undef EPS
#undef JMAX
/* (C) Copr. 1986-92 Numerical Recipes Software 71.+I0>+. */
