#include<stdio.h>
#include<stdlib.h>
#include<string.h>
#include<mpi.h>
#include "pmheader.h"
#include <hdf5.h>

#define HEADERSETNAME "HeaderOfStructure"
#define DATASETNAME "DataOfStructure"
typedef struct _header {
	int ng,ngz,nspace,iseed;
	float size,omep,omeplam,amax,astep,pconv,vconv;
	float a,redshift;
	int np,myrank,nrank,nred10;
} header;
typedef struct _Menu_item{
    char *title;
    int offset;
    char *type;
} Menu_item;
Menu_item sim_data[] = {
	/*
	{"mass(1./total #)",HOFFSET(basicparticletype,mass),"H5T_NATIVE_FLOAT"},
	*/
	{"x(in mesh scale)",HOFFSET(basicparticletype,x),"H5T_NATIVE_FLOAT"},
	{"y(in mesh scale)",HOFFSET(basicparticletype,y),"H5T_NATIVE_FLOAT"},
	{"z(in mesh scale)",HOFFSET(basicparticletype,z),"H5T_NATIVE_FLOAT"},
	{"vx(in mesh scale)",HOFFSET(basicparticletype,vx),"H5T_NATIVE_FLOAT"},
	{"vy(in mesh scale)",HOFFSET(basicparticletype,vy),"H5T_NATIVE_FLOAT"},
	{"vz(in mesh scale)",HOFFSET(basicparticletype,vz),"H5T_NATIVE_FLOAT"},
	/*
	{"freework",HOFFSET(basicparticletype,freework),"H5T_NATIVE_FLOAT"},
	{"epssq",HOFFSET(basicparticletype,epssq),"H5T_NATIVE_FLOAT"},
	*/
	NULL
};
Menu_item sim_inform[]={
	{"ng",HOFFSET(header,ng),"H5T_NATIVE_INT"},
	{"ngz",HOFFSET(header,ngz),"H5T_NATIVE_INT"},
	{"nspace",HOFFSET(header,nspace),"H5T_NATIVE_INT"},
	{"iseed",HOFFSET(header,iseed),"H5T_NATIVE_INT"},
	{"size",HOFFSET(header,size),"H5T_NATIVE_FLOAT"},
	{"omep",HOFFSET(header,omep),"H5T_NATIVE_FLOAT"},
	{"omeplam",HOFFSET(header,omeplam),"H5T_NATIVE_FLOAT"},
	{"amax",HOFFSET(header,amax),"H5T_NATIVE_FLOAT"},
	{"astep",HOFFSET(header,astep),"H5T_NATIVE_FLOAT"},
	{"pconv",HOFFSET(header,pconv),"H5T_NATIVE_FLOAT"},
	{"vconv",HOFFSET(header,vconv),"H5T_NATIVE_FLOAT"},
	{"a",HOFFSET(header,a),"H5T_NATIVE_FLOAT"},
	{"redshift",HOFFSET(header,redshift),"H5T_NATIVE_FLOAT"},
	{"np",HOFFSET(header,np),"H5T_NATIVE_INT"},
	{"myrank",HOFFSET(header,myrank),"H5T_NATIVE_INT"},
	{"nrank",HOFFSET(header,nrank),"H5T_NATIVE_INT"},
	{"nred10",HOFFSET(header,nred10),"H5T_NATIVE_INT"},
	NULL
};
void hdf5_insert(hid_t id,Menu_item *);
int wdata2disk(int ng,int ngz,int nspace,int iseed,float size,float omep,
		float omeplam, float amax,float astep,float pconv,float vconv,float a,
		float redshift,basicparticletype *particles, int np,int myrank,
		int nrank,int nred10){
	FILE *fp;
	int i,j,k;
	char *fileheader = FILEHEADER;
	char procname[3],number[3];
	/*  setting for HDF5 file format */
	header file_inform;
	hsize_t dim;
	hid_t file,header_set,data_set,header_space,data_space;
	hid_t group_id,group_id_info,group_id_data;
	hid_t header_id,data_id;
	herr_t status;
	char *filename = (char *)malloc(sizeof(char)*50);
	sprintf(procname,"%03d",myrank);
	sprintf(number,"%03d",nred10);
	strcpy(filename,fileheader);
	strcat(filename,number);
	strcat(filename,"_");
	strcat(filename,procname);
	strcat(filename,".hdf5");

	file_inform.ng = ng;
	file_inform.ngz = ngz;
	file_inform.nspace = nspace;
	file_inform.iseed = iseed;
	file_inform.size = size;
	file_inform.omep = omep;
	file_inform.omeplam = omeplam;
	file_inform.amax = amax;
	file_inform.astep = astep;
	file_inform.pconv = pconv;
	file_inform.vconv = vconv;
	file_inform.a = a;
	file_inform.redshift = redshift;
	file_inform.np = np;
	file_inform.myrank = myrank;
	file_inform.nrank = nrank;
	file_inform.nred10 = nred10;

	if(myrank == 0) 
		fprintf(stdout," writing data at a=%g,z=%g \n",a,redshift);
	for(i=0;i<nrank;i++){
		if(myrank == i){
			file = H5Fcreate(filename,H5F_ACC_TRUNC, H5P_DEFAULT, H5P_DEFAULT);
			group_id = H5Gcreate(file,"/SIMULATION",0);

#ifdef INCLUDE_TREE_FORCE
			H5Gset_comment(group_id,".","PM/TREE");
#else 
			H5Gset_comment(group_id,".","PM_ONLY");
#endif
			/*    */
			header_id = H5Tcreate (H5T_COMPOUND,sizeof(header));
			hdf5_insert(header_id,sim_inform);
			dim = 1;
			header_space = H5Screate_simple(1,&dim,NULL);
			group_id_info = H5Gcreate(file,"/SIMULATION/INFORMATION",0);
			header_set = H5Dcreate(group_id_info,"SIM_PARAMETERS",header_id,
					header_space,H5P_DEFAULT);
			status = H5Dwrite(header_set,header_id,H5S_ALL, H5S_ALL, 
					H5P_DEFAULT,&file_inform);
			H5Dclose(header_set);H5Tclose(header_id); H5Sclose(header_space);
			H5Gclose(group_id_info);

			/* from this  -> data */
			data_id = H5Tcreate (H5T_COMPOUND,sizeof(basicparticletype));
			hdf5_insert(data_id,sim_data);
			dim = np;
			data_space = H5Screate_simple(1,&dim,NULL);
			data_set = H5Dcreate(file,"/SIMULATION/DATA",data_id,data_space,
					H5P_DEFAULT);
			status = H5Dwrite(data_set,data_id,H5S_ALL, H5S_ALL,
					H5P_DEFAULT,particles);
			H5Dclose(data_set);H5Tclose(data_id);H5Sclose(data_space);
			/*
			H5Dclose(group_id_data);
			*/
			/*    */

			H5Gclose(group_id);
			H5Fclose(file);
		}
		MPI_Barrier(MPI_COMM_WORLD);
	}
	if(myrank == 0) 
		fprintf(stdout," well wrote data at a=%g,z=%g \n",a,redshift);
	return 0;
}

int c_rdatafromdisk(int *ng,int *ngz,int *nspace,int *iseed,
	        float *size,float *omep,float *omeplam,
	        float *amax,float *astep,float *pconv,float *vconv,float *a,
	        float *redshift,float *ken,basicparticletype *particles,
	        int *np,int myrank,int nrank,int inumber,int nstep){
	FILE *fp;
	int i,j,k,nmax,ptrsize();
	char *fileheader = FILEHEADER;
	char procname[3],number[3];
	/*  setting for HDF5 file format */
	header file_inform;
	hsize_t dim;
	hid_t file,header_set,data_set,header_space,data_space;
	hid_t group_id,group_id_info,group_id_data;
	hid_t header_id,data_id;
	herr_t status;
	char *filename = (char *)malloc(sizeof(char)*50);
	sprintf(procname,"%03d",myrank);
	sprintf(number,"%03d",inumber);
	strcpy(filename,fileheader);
	strcat(filename,number);
	strcat(filename,"_");
	strcat(filename,procname);
	strcat(filename,".hdf5");
	if(myrank == 0) 
		fprintf(stdout," reading data  \n");
	for(i=0;i<nrank;i++){
		if(myrank == i){
			file = H5Fopen(filename,H5F_ACC_RDONLY, H5P_DEFAULT);
			header_set = H5Dopen(file,"/SIMULATION/INFORMATION/SIM_PARAMETERS");
			header_id = H5Tcreate (H5T_COMPOUND,sizeof(header));
			hdf5_insert(header_id,sim_inform);
			dim = 1;
			status = H5Dread(header_set,header_id,H5S_ALL, H5S_ALL, 
					H5P_DEFAULT,&file_inform);
			H5Dclose(header_set);H5Tclose(header_id);
			*ng = file_inform.ng;
			*ngz = file_inform.ngz;
			*nspace = file_inform.nspace;
			*iseed = file_inform.iseed;
			*size = file_inform.size;
			*omep = file_inform.omep;
			*omeplam = file_inform.omeplam;
			*amax = file_inform.amax;
			*astep = file_inform.astep;
			*pconv = file_inform.pconv;
			*vconv = file_inform.vconv;
			*a = file_inform.a;
			*redshift = file_inform.redshift;
			*np = file_inform.np;
			nmax = ptrsize(particles)/sizeof(basicparticletype);
			if( nmax <  *np) {
				fprintf(stderr,
						"%d(%d) Not enough memory for %d particles in pm()\n",
						myrank,nmax,*np);
				exit(0);
			}
			data_set = H5Dopen(file,"/SIMULATION/DATA");
			data_id = H5Tcreate (H5T_COMPOUND,sizeof(basicparticletype));
			hdf5_insert(data_id,sim_data);
			dim = *np;
			status = H5Dread(data_set,data_id,H5S_ALL, H5S_ALL, 
					H5P_DEFAULT,particles);
			H5Dclose(data_set);H5Tclose(data_id);
			H5Fclose(file);
			for(j=0;j<(*np);j++){
				particles[j].indx = (*np) *myrank+j;
			}
		}
		MPI_Barrier(MPI_COMM_WORLD);
	}
	if(myrank == 0) 
		fprintf(stdout," well read data at a=%g,z=%g \n",a,redshift);
	return 0;
}

void hdf5_insert(hid_t id,Menu_item *menu){
	int i=0;
	hid_t type;
	while(menu[i].title != NULL){
		if(strcmp(menu[i].type,"H5T_NATIVE_INT")==0) 
			type = H5T_NATIVE_INT;
		else if(strcmp(menu[i].type,"H5T_NATIVE_FLOAT")==0)
			type = H5T_NATIVE_FLOAT;
		else if(strcmp(menu[i].type,"H5T_NATIVE_CHAR")==0) 
			type = H5T_NATIVE_CHAR;
		else if(strcmp(menu[i].type,"H5T_NATIVE_LONG")==0)
			type = H5T_NATIVE_LONG;
		else if(strcmp(menu[i].type,"H5T_NATIVE_UINT")==0)
			type = H5T_NATIVE_UINT;
		else if(strcmp(menu[i].type,"H5T_NATIVE_DOUBLE")==0) 
			type = H5T_NATIVE_DOUBLE;
		else {
			fprintf(stderr,"%s : ",menu[i].type);
			fprintf(stderr,"Not supported data type in HDF format\n");
			fprintf(stderr,"Exiting \n");
			MPI_Barrier(MPI_COMM_WORLD);
			MPI_Finalize();
			exit(0);
		}
		H5Tinsert(id,menu[i].title,menu[i].offset,type);
		i = i + 1;
	}
}
