#include "defs.h"
#include "extern.h"

float height[512];
float z0n[512], z1n[512], dzn[512];
int senddownlist[20], senduplist[20], recvlist[20];
MPI_Request req1[20], req2[20], req3[20], req4[20];
MPI_Status status1[20], status2[20], status3[20], status4[20];
int scount[20], rcount[20];
basicparticletype *psend[20], *precv[20], *buffer;

extern float zboxup,zboxsize;

int GetVisitors(Np,BasicParticles,dz,z0,height)
int Np;
basicparticletype *BasicParticles;
float dz, z0, height;
{
	int i, j, k, p, s;
	int senddowncount, sendupcount, recvcount;
	INT8 Nmax;
	int nvisitors;
	int offset; 
        int buffersize;
	float z1, z2, dz1, zmin, zmax; 
	double zmid, dz2;
	basicparticletype *bp;

    Nmax = Np + freespace()/sizeof(basicparticletype);
    BasicParticles = (basicparticletype *)
        resizelast(BasicParticles,Nmax*sizeof(basicparticletype));

	z1 = z0 + height;
/*
 * Distribute information on slice widths
 */

    (void) MPI_Gather(&dz,1,MPI_FLOAT,dzn,1,MPI_FLOAT,0,MPI_COMM_WORLD);
	(void) MPI_Bcast(dzn,numprocs,MPI_FLOAT,0,MPI_COMM_WORLD);
    (void) MPI_Gather(&z0,1,MPI_FLOAT,z0n,1,MPI_FLOAT,0,MPI_COMM_WORLD);
	(void) MPI_Bcast(z0n,numprocs,MPI_FLOAT,0,MPI_COMM_WORLD);
    (void) MPI_Gather(&z1,1,MPI_FLOAT,z1n,1,MPI_FLOAT,0,MPI_COMM_WORLD);
	(void) MPI_Bcast(z1n,numprocs,MPI_INT,0,MPI_COMM_WORLD);
#ifdef DEBUG
	fprintf(stderr,"P%d have passed the gathering of slice widths from others\n",myid);
#endif

/*
 * Search up for processors to receive data
 */
	recvcount = 0;
	z2 = z1n[myid]+dzn[myid];
	for(i=myid+1; i<=myid+numprocs/2; i++) {
		p = (i+numprocs) % numprocs;
		dz1 = z2 - z0n[p]; if ( dz1 > zboxsize*0.5 ) dz1 -= zboxsize;
		if(  dz1 > 0.0 ) {
			recvlist[recvcount] = p;
			recvcount++;
		}
		else 
			break;
	}
/*
 * Search down for processors to receive data
 */
	z2 = z0n[myid]-dzn[myid];
	for(i=myid-1; i>=myid-numprocs/2; i--) {
		p = (i+numprocs) % numprocs;
		dz1 = z2 - z1n[p]; if ( dz1 < -zboxsize*0.5 ) dz1 += zboxsize;
		if( dz1 < 0.0 ) {
			recvlist[recvcount] = p;
			recvcount++;
		}
		else 
			break;
	}
/*
 * Search up for processors to send data
 */
	sendupcount = 0;
	for(i=myid+1; i<=myid+numprocs/2; i++) {
		p = (i+numprocs) % numprocs;
		z2 = z0n[p] - dzn[p];
		dz1 = z2 - z1n[myid]; if ( dz1 < -zboxsize*0.5 ) dz1 += zboxsize;
		if(  dz1 < 0.0 ) {
			senduplist[sendupcount] = p;
			sendupcount++;
		}
		else 
			break;
	}
/*
 * Search down for processors to send data
 */
	senddowncount = 0;
	for(i=myid-1; i>=myid-numprocs/2; i--) {
		p = (i+numprocs) % numprocs;
		z2 = z1n[p] + dzn[p];
		dz1 = z2 - z0n[myid]; if ( dz1 > zboxsize*0.5 ) dz1 -= zboxsize;
		if(  dz1 > 0.0 ) {
			senddownlist[senddowncount] = p;
			senddowncount++;
		}
		else 
			break;
	}

/* Prepare send lists */
	psend[0] = BasicParticles + Np;
	s = 0;
	offset = 0;
	for(i=0; i<senddowncount; i++) {
		p = senddownlist[i];
		z2 = z1n[p] + dzn[p];
		scount[s] = 0;
		for(j=0; j<Np; j++) {
			dz1 = z2 - BasicParticles[j].z; if ( dz1 > zboxsize*0.5 ) dz1 -= zboxsize;
			if( dz1 > 0.0 ) {
				psend[s][scount[s]] = BasicParticles[j];
				scount[s]++;
			}
		}
		psend[s+1] = psend[s] + scount[s];
		offset += scount[s];
		s++;
	}
	for(i=0; i<sendupcount; i++) {
		p = senduplist[i];
		z2 = z0n[p] - dzn[p];
		scount[s] = 0;
		for(j=0; j<Np; j++) {
			dz1 = z2 - BasicParticles[j].z; if ( dz1 < -0.5*zboxsize ) dz1 += zboxsize;
			if( dz1 < 0.0 ) {
				psend[s][scount[s]] = BasicParticles[j];
				scount[s]++;
			}
		}
		psend[s+1] = psend[s] + scount[s];
		offset += scount[s];
		s++;
	}

	s = 0;
	for(i=0; i<senddowncount; i++) {
		p = senddownlist[i];
		MPI_Isend(scount+s, 1, MPI_INT, p, VCOUNT, MPI_COMM_WORLD, req3+s);
		MPI_Wait(req3+s,status3+s);
#ifdef DEBUG
fprintf(stderr,"P%d sending %d to P%d\n",myid,scount[s],p);
#endif
		s++;
	}
	for(i=0; i<sendupcount; i++) {
		p = senduplist[i];
		MPI_Isend(scount+s, 1, MPI_INT, p, VCOUNT, MPI_COMM_WORLD,req3+s);
		MPI_Wait(req3+s,status3+s);
#ifdef DEBUG
fprintf(stderr,"P%d sending %d to P%d\n",myid,scount[s],p);
#endif
		s++;
	}

/* Complete receives */
	precv[0] = BasicParticles+Np+offset;
	for(i=0; i<recvcount; i++) {
		p = recvlist[i];
		MPI_Irecv(rcount+i, 1, MPI_INT, p, VCOUNT, MPI_COMM_WORLD, req1+i);
		MPI_Wait(req1+i,status1+i);

		precv[i+1] = precv[i] + rcount[i];
		offset += rcount[i];
#ifdef DEBUG
fprintf(stderr,"P%d receiving %d from %d\n",myid,rcount[i],p);
#endif
	}

    buffersize = MPI_BSEND_OVERHEAD;
    for(i=0; i<s; i++) {
        buffersize += (scount[i]*sizeof(basicparticletype) 
			+ MPI_BSEND_OVERHEAD);
    }
    if( buffersize > (Nmax - Np - offset)*sizeof(basicparticletype) ) {
        fprintf(stderr,"OUT OF MEMORY buffersize=%lld , Nmax=%lld - EXITING\n",buffersize,
		(Nmax-Np-offset)*sizeof(basicparticletype));
        fprintf(stderr,"OUT OF MEMORY - EXITING\n");
        EXIT(0); /*this does an MPI_Abort */
    }

	buffer = BasicParticles+Np+offset;
	MPI_Buffer_attach(buffer, buffersize);

	s = 0;
	for(i=0; i<senddowncount; i++) {
		p = senddownlist[i];
		MPI_Ibsend(psend[s], scount[s]*sizeof(basicparticletype), 
			MPI_BYTE, p, VPARTICLES, MPI_COMM_WORLD, req4+s);
		MPI_Wait(req4+s,status4+s);
		s++;
	}
	for(i=0; i<sendupcount; i++) {
		p = senduplist[i];
		MPI_Ibsend(psend[s], scount[s]*sizeof(basicparticletype), 
			MPI_BYTE, p, VPARTICLES, MPI_COMM_WORLD, req4+s);
		MPI_Wait(req4+s,status4+s);
		s++;
	}

/* Complete receives */
	for(i=0; i<recvcount; i++) {
		p = recvlist[i];
		MPI_Irecv(precv[i], rcount[i]*sizeof(basicparticletype), 
			MPI_BYTE, p, VPARTICLES, MPI_COMM_WORLD, req2+i);
		MPI_Wait(req2+i,status2+i);
	}
	(void) MPI_Buffer_detach(buffer, &buffersize);

/* recompactify the BasicParticles array with the received particle list */
	k = Np;
	for(i=0; i<recvcount; i++) {
		for(j=0; j<rcount[i]; j++) {
			BasicParticles[k] = precv[i][j];
			k++;
		}
	}

	zmid = (double) (0.5*(z0 + z1));
	zmin = 1.0e30; zmax = -1.0e30;
	for(i=0; i<k; i++) {
		bp = &BasicParticles[i];
		dz2 = (double) (bp->z) - zmid;
		if( dz2 < -0.5*zboxsize) dz2 += zboxsize;
		else if( dz2 > zboxsize*0.5) dz2 -= zboxsize;
		bp->z = (float) (dz2 + zmid);
		if( dz2 < zmin ) zmin = dz2;
		if( dz2 > zmax ) zmax = dz2;
	}
#ifdef DEBUG1
	fprintf(stderr,"P%d k %d z0 %g z1 %g zmid %g zmin %g zmax %g\n",
		myid,k,z0,z1,zmid,zmin+zmid,zmax+zmid);
#endif

/* Free up the memory at the end of the list for other purposes */
    BasicParticles = (basicparticletype *)
        resizelast(BasicParticles, k*sizeof(basicparticletype));

	nvisitors = k - Np;
	SYNCHRONIZE();

	return nvisitors;
}
