#include "defs.h"
#include "extern.h"


#define no_DEBUG1

#define MAX_SLICES 512
#define MAX_REQS 20

float height[MAX_SLICES];
float z0n[MAX_SLICES], z1n[MAX_SLICES], dzn[MAX_SLICES];
int senddownlist[MAX_REQS], senduplist[MAX_REQS], recvlist[MAX_REQS];
MPI_Request req1[MAX_REQS], req2[MAX_REQS], req3[MAX_REQS], req4[MAX_REQS];
MPI_Status status1[MAX_REQS], status2[MAX_REQS], status3[MAX_REQS], status4[MAX_REQS];
int scount[MAX_REQS], rcount[MAX_REQS];
basicparticletype *psend[MAX_REQS], *precv[MAX_REQS], *buffer;

int GetVisitors(Np,BasicParticles,dz,z0,height)
int Np;
basicparticletype *BasicParticles;
float dz, z0, height;
{
        int i, j, k, p, s;
        int senddowncount, sendupcount, recvcount;
        int Nmax;
        int nvisitors;
        int offset, buffersize;
        float z1, z2, dz1, zmin, zmax; 
        double zmid, dz2;
        basicparticletype *bp;

        if ( numprocs == 1 ) return 0; /* no visitors */

        Nmax = Np + freespace()/sizeof(basicparticletype);
        BasicParticles = (basicparticletype *)
        resizelast(BasicParticles,Nmax*sizeof(basicparticletype));

        if ( numprocs > MAX_SLICES )
        {
                fprintf(stderr,"getvisitors: increase MAX_SLICES to at least %d or run on less nodes - currently %d\n", numprocs, MAX_SLICES );
                EXIT(0); /*this does an MPI_Abort */
        }

        z1 = z0 + height;
/*
 * Distribute information on slice widths
 */

        (void) MPI_Gather(&dz,1,MPI_FLOAT,dzn,1,MPI_FLOAT,0,MPI_COMM_WORLD);
        (void) MPI_Bcast(dzn,numprocs,MPI_FLOAT,0,MPI_COMM_WORLD);
        (void) MPI_Gather(&z0,1,MPI_FLOAT,z0n,1,MPI_FLOAT,0,MPI_COMM_WORLD);
        (void) MPI_Bcast(z0n,numprocs,MPI_FLOAT,0,MPI_COMM_WORLD);
        (void) MPI_Gather(&z1,1,MPI_FLOAT,z1n,1,MPI_FLOAT,0,MPI_COMM_WORLD);
        (void) MPI_Bcast(z1n,numprocs,MPI_INT,0,MPI_COMM_WORLD);

/*
 * Search up for processors to receive data
 */
        recvcount = 0;
        z2 = z1n[myid]+dzn[myid];
        for(i=myid+1; i<=myid+numprocs/2; i++) {
                p = (i+numprocs) % numprocs;
                dz1 = z2 - z0n[p]; if ( dz1 > 0.5 ) dz1 -= 1.0;
                if(  dz1 > 0.0 ) {
#ifdef DEBUG1
                        fprintf( stderr, "p%d getvisitors: i %d recv up   loop from p%d - yup %g>0\n", myid, i, p, dz1 );
#endif
                        recvlist[recvcount] = p;
                        recvcount++;
                }
                else {
#ifdef DEBUG1
                        fprintf( stderr, "p%d getvisitors: i %d recv up   loop to p%d - nope %g\n", myid, i, p, dz1 );
#endif
                        break;
                }
        }
/*
 * Search down for processors to receive data
 */
        z2 = z0n[myid]-dzn[myid];
        for(i=myid-1; i>=myid-numprocs/2; i--) {
                p = (i+numprocs) % numprocs;
                dz1 = z2 - z1n[p]; if ( dz1 < -0.5 ) dz1 += 1.0;
                if( dz1 < 0.0 ) {
#ifdef DEBUG1
                        fprintf( stderr, "p%d getvisitors: i %d recv down loop from p%d - yup %g<0\n", myid, i, p, dz1 );
#endif
                        recvlist[recvcount] = p;
                        recvcount++;
                }
                else {
#ifdef DEBUG1
                        fprintf( stderr, "p%d getvisitors: i %d recv down loop to p%d - nope %g\n", myid, i, p, dz1 );
#endif
                        break;
                }
        }
/*
 * Search up for processors to send data
 */
        sendupcount = 0;
        for(i=myid+1; i<=myid+numprocs/2; i++) {
                p = (i+numprocs) % numprocs;
                z2 = z0n[p] - dzn[p];
                dz1 = z2 - z1n[myid]; if ( dz1 < -0.5 ) dz1 += 1.0;
                if(  dz1 < 0.0 ) {
#ifdef DEBUG1
                        fprintf( stderr, "p%d getvisitors: i %d send up   loop to p%d - yup %g<0\n", myid, i, p, dz1 );
#endif
                        senduplist[sendupcount] = p;
                        sendupcount++;
                }
                else {
#ifdef DEBUG1
                        fprintf( stderr, "p%d getvisitors: i %d send up   loop to p%d - nope %g\n", myid, i, p, dz1 );
#endif
                        break;
                }
        }
/*
 * Search down for processors to send data
 */
        senddowncount = 0;
        for(i=myid-1; i>=myid-numprocs/2; i--) {
                p = (i+numprocs) % numprocs;
                z2 = z1n[p] + dzn[p];
                dz1 = z2 - z0n[myid]; if ( dz1 > 0.5 ) dz1 -= 1.0;
                if(  dz1 > 0.0 ) {
#ifdef DEBUG1
                        fprintf( stderr, "p%d getvisitors: i %d send down loop to p%d - yup %g>0\n", myid, i, p, dz1 );
#endif
                        senddownlist[senddowncount] = p;
                        senddowncount++;
                }
                else {
#ifdef DEBUG1
                        fprintf( stderr, "p%d getvisitors: i %d send down loop to p%d - nope %g\n", myid, i, p, dz1 );
#endif
                        break;
                }
        }

#ifdef DEBUG1
        fprintf( stderr, "p%d getvisitors: senddown %d, up %d, recv %d\n", myid, senddowncount, sendupcount, recvcount );
#endif

/* Prepare send lists */
        psend[0] = BasicParticles + Np;
        s = 0;
        offset = 0;
        for(i=0; i<senddowncount; i++) {
                p = senddownlist[i];
                z2 = z1n[p] + dzn[p];
                scount[s] = 0;
                for(j=0; j<Np; j++) {
                        dz1 = z2 - BasicParticles[j].z; if ( dz1 > 0.5 ) dz1 -= 1.0;
                        if( dz1 > 0.0 ) {
                                psend[s][scount[s]] = BasicParticles[j];
                                scount[s]++;
                        }
                }
                psend[s+1] = psend[s] + scount[s];
                offset += scount[s];
                s++;
        }
        for(i=0; i<sendupcount; i++) {
                p = senduplist[i];
                z2 = z0n[p] - dzn[p];
                scount[s] = 0;
                for(j=0; j<Np; j++) {
                        dz1 = z2 - BasicParticles[j].z; if ( dz1 < -0.5 ) dz1 += 1.0;
                        if( dz1 < 0.0 ) {
                                psend[s][scount[s]] = BasicParticles[j];
                                scount[s]++;
                        }
                }
                psend[s+1] = psend[s] + scount[s];
                offset += scount[s];
                s++;
        }

        if ( senddowncount + sendupcount + recvcount > MAX_REQS )
        {
                fprintf(stderr,"getvisitors: increase the amount of simultaneous sends to at least %d - currently %d\n", senddowncount + sendupcount + recvcount, MAX_REQS );
                EXIT(0); /*this does an MPI_Abort */
        }

        s = 0;
    /* Post receives */
        for(i=0; i<recvcount; i++) {
                p = recvlist[i];
                MPI_Irecv(rcount+i, 1, MPI_INT, p, VCOUNT, MPI_COMM_WORLD, req3+s);
                s++;
        }

        /* Do sends */
        for(i=0; i<senddowncount; i++) {
                p = senddownlist[i];
                MPI_Isend(scount+s, 1, MPI_INT, p, VCOUNT, MPI_COMM_WORLD, req3+s);
#ifdef DEBUG1
                fprintf(stderr,"P%d down sending %d to P%d\n",myid,scount[s],p);
#endif
                s++;
        }
        for(i=0; i<sendupcount; i++) {
                p = senduplist[i];
                MPI_Isend(scount+s, 1, MPI_INT, p, VCOUNT, MPI_COMM_WORLD,req3+s);
#ifdef DEBUG1
                fprintf(stderr,"P%d up   sending %d to P%d\n",myid,scount[s],p);
#endif
                s++;
        }
#ifdef DEBUG1
        fprintf( stderr, "p%d getvisitors: offset %d\n", myid, offset );
#endif

        MPI_Waitall( s, req3, status3 );

    /* buffersize = 3*offset*sizeof(basicparticletype); */
    buffersize = 0;
        for(i=0; i<s; i++) {
                buffersize += (scount[i]*sizeof(basicparticletype) + MPI_BSEND_OVERHEAD);
        }
        if( buffersize > (Nmax - Np - offset)*sizeof(basicparticletype) ) {
                fprintf(stderr,"OUT OF MEMORY - EXITING\n");
                EXIT(0); /*this does an MPI_Abort */
        }

        /* use recv'd data */
        precv[0] = BasicParticles+Np+offset;
        for(i=0; i<recvcount; i++) {
                p = recvlist[i];
                precv[i+1] = precv[i] + rcount[i];
                offset += rcount[i];
#ifdef DEBUG1
                fprintf(stderr,"P%d receiving %d from %d\n",myid,rcount[i],p);
#endif
        }

        buffer = BasicParticles+Np+offset;
#ifdef DEBUG1
        fprintf( stderr, "p%d getvisitors: buffer %p, buffersize %d\n", myid, buffer, buffersize );
#endif
        MPI_Buffer_attach(buffer, buffersize);

        s = 0;
    /* Post receives */
        for(i=0; i<recvcount; i++) {
                p = recvlist[i];
                MPI_Irecv(precv[i], rcount[i]*sizeof(basicparticletype), 
                        MPI_BYTE, p, VPARTICLES, MPI_COMM_WORLD, req4+s);
                s++;
        }

        /* Do sends */
        for(i=0; i<senddowncount; i++) {
                p = senddownlist[i];
                MPI_Ibsend(psend[s], scount[s]*sizeof(basicparticletype), 
                        MPI_BYTE, p, VPARTICLES, MPI_COMM_WORLD, req4+s);
                s++;
        }
        for(i=0; i<sendupcount; i++) {
                p = senduplist[i];
                MPI_Ibsend(psend[s], scount[s]*sizeof(basicparticletype), 
                        MPI_BYTE, p, VPARTICLES, MPI_COMM_WORLD, req4+s);
                s++;
        }

        MPI_Waitall( s, req4, status4 );
        (void) MPI_Buffer_detach(buffer, &buffersize);

        k = Np;
        for(i=0; i<recvcount; i++) {
                for(j=0; j<rcount[i]; j++) {
                        BasicParticles[k] = precv[i][j];
                        k++;
                }
        }

        zmid = (double) (0.5*(z0 + z1));
        zmin = 1.0e30; zmax = -1.0e30;
        for(i=0; i<k; i++) {
                bp = &BasicParticles[i];
                dz2 = (double) (bp->z) - zmid;
                if( dz2 < -0.5) dz2 += 1.0;
                if( dz2 > 0.5) dz2 -= 1.0;
                bp->z = (float) (dz2 + zmid);
                if( dz2 < zmin ) zmin = dz2;
                if( dz2 > zmax ) zmax = dz2;
        }
#ifdef DEBUG1
        fprintf(stderr,"P%d k %d z0 %g z1 %g zmid %g zmin %g zmax %g\n",
                myid,k,z0,z1,zmid,zmin+zmid,zmax+zmid);
#endif

    BasicParticles = (basicparticletype *)
        resizelast(BasicParticles, k*sizeof(basicparticletype));

        nvisitors = k - Np;
        SYNCHRONIZE();

        return nvisitors;
}
