#include "defs.h"
#include "extern.h"
#include "extern_simparam.h"

#define onethird 0.333333333
#define qxx0 (c->qxx)
#define qyy0 (c->qyy)
#define qzz0 (c->qzz)
#define qxy0 (c->qxy)
#define qxz0 (c->qxz)
#define qyz0 (c->qyz)
#define mxy qxy0
#define myx qxy0
#define mxz qxz0
#define mzx qxz0
#define myz qyz0
#define mzy qyz0
#define momnode (c->mom)
#define LIST_SIZE 1024

#define A1 (1.0/(256.*256.))
#define A2 (0.78224E+00)
#define A3 (0.37971E-06)
#define A4 (-0.60338E+00)
#define A5 (-0.34419E-07)
#define A6 (-0.39741E+01)
#define A7 (-0.10607E+01)
#define A8 (-0.38145E+00)

#define NPOINTS 100000

extern float yboxup,yboxsize,zboxup,zboxsize;

#define COPYCELL() \
{ \
    clist[ccount].type = c->type; \
    clist[ccount].mass = c->mass; \
    clist[ccount].r    = c->r; \
    clist[ccount].epssq  = c->epssq; \
    if( c->type == CELL_TYPE ) { \
        clist[ccount].qxx  = qxx0; \
        clist[ccount].qyy  = qyy0; \
        clist[ccount].qzz  = qzz0; \
        clist[ccount].qxy  = qxy0; \
        clist[ccount].qxz  = qxz0; \
        clist[ccount].qyz  = qyz0; \
        clist[ccount].mom  = momnode; \
    } \
}

#define COPYPARTICLE() \
{ \
    plist[pcount].type = c->type; \
    plist[pcount].mass = c->mass; \
    plist[pcount].r    = c->r; \
    plist[pcount].epssq  = c->epssq; \
}


#define CORRECTION(rad2) \
{ \
	int i; \
	i = (int) (rad2/cordr2); \
	if( i > NPOINTS-1 ) \
		pmforce = L2/rad2; \
	pmforce = (float) (ca0[i] + ma0[i]*(rad2 - ra2[i]));\
}

#define QUADCORRECTION(rad2) \
{ \
	int i; \
	i = (int) (rad2/cordr2); \
	if( i > NPOINTS-1 ) { \
		pmf1 = -3*L2/rad2; pmf2 = 15*L2/rad2; \
	} \
	pmf1 = (float) (ca1[i] + ma1[i]*(rad2 - ra2[i]));\
	pmf2 = (float) (ca2[i] + ma2[i]*(rad2 - ra2[i]));\
}

#define OPENINGCRITERION() \
{ \
	coordStruct r; \
	coordStruct lcorner, rcorner, length, drc; \
	float dr2; \
 \
	r = c->r; \
	lcorner = theGroupCell.corner; \
	length = theGroupCell.length; \
	rcorner.x = lcorner.x + length.x; \
	rcorner.y = lcorner.y + length.y; \
	rcorner.z = lcorner.z + length.z; \
	 \
 \
	drc.x = drc.y = drc.z = 0; \
    if( r.x < lcorner.x || r.x > rcorner.x ) { \
        dx1 = fabs(lcorner.x - r.x); \
        if( dx1 > 0.5 ) dx1 = 1.0 - dx1; \
        dx2 = fabs(rcorner.x - r.x); \
        if( dx2 > 0.5 ) dx2 = 1.0 - dx2; \
        drc.x = dx2; \
        if( dx1 < dx2 ) \
            drc.x = dx1; \
    } \
    if( r.y < lcorner.y || r.y > rcorner.y ) { \
        dy1 = fabs(lcorner.y - r.y); \
        if( dy1 > 0.5*yboxsize ) dy1 = yboxsize - dy1; \
        dy2 = fabs(rcorner.y - r.y); \
        if( dy2 > 0.5*yboxsize ) dy2 = yboxsize - dy2; \
        drc.y = dy2; \
        if( dy1 < dy2 ) \
            drc.y = dy1; \
    } \
    if( r.z < lcorner.z || r.z > rcorner.z ) { \
        dz1 = fabs(lcorner.z - r.z); \
        if( dz1 > 0.5*zboxsize ) dz1 = zboxsize - dz1; \
        dz2 = fabs(rcorner.z - r.z); \
        if( dz2 > 0.5*zboxsize ) dz2 = zboxsize - dz2; \
        drc.z = dz2; \
        if( dz1 < dz2 ) \
            drc.z = dz1; \
    } \
	dr2 = drc.x*drc.x + drc.y*drc.y + drc.z*drc.z; \
	OpeningDecision =  (dr2 < c->cellsq); \
}


static coordStruct corner, length, mid, drc, drg;
static cellStruct plist[LIST_SIZE], clist[LIST_SIZE];
static cellStruct theGroupCell;
static float x, y, z;
static float x2, y2, z2, epssq;
static float dr2, sr, sr2, phii, mor3, phisum;
static float or5, temp, phiquad;
static coordStruct acc0, pos0;
static particleStruct *p;
static cellStruct *c;


float cordr2; 
float pmforce, pmf1, pmf2;
static double ra2[NPOINTS]; 
static double ca0[NPOINTS], ma0[NPOINTS];
static double ca1[NPOINTS], ma1[NPOINTS];
static double ca2[NPOINTS], ma2[NPOINTS];
extern float EPSSQ;

InitCorrectionFunction()
{
	int i;
	double x, dx;
	double xm1, xm2, xp1, xp2;
	double jphi(), jphix();
	double pmf(), qpmf(), pmf0, pmf1, pmf2;
	FILE *fcorfile;

	if( myid == 0 ) {
		fcorfile = fopen("fcor.dat","w");
	}
	cordr2 = (16.0*SPHERE_RADIUS*SPHERE_RADIUS)/NPOINTS;
	ra2[0] = 0.0; ca0[0] = 0.0;
	ca1[0] = ca2[0] = 0;
	for(i=1; i<NPOINTS; i++) {
		ra2[i] = i*cordr2;

		dx = 0.5*(sqrt((i+1)*cordr2) - sqrt(i*cordr2));
		x = sqrt(ra2[i]);
		xm1 = x - dx; xm2 = x - 2*dx; xp1 = x + dx; xp2 = x + 2*dx; 

		pmf0 = pmf(x);
		pmf1 = (qpmf(xp1) - qpmf(xm1))/(2.0*dx);
		pmf2 = (qpmf(xp2) - 2*qpmf(x) + qpmf(xm2))/(4.0*dx*dx);

		ca0[i] = pmf0;
		ca1[i] = x*pmf1 - qpmf(x);
		ca2[i] = 3*qpmf(x) +x*( -3*pmf1 + x*pmf2);

		ra2[i] /= L2;
		if( myid == 0 ) {
			if( i % 100 == 0 ) {
			fprintf(fcorfile,"%.8g %.8g %.8g %.8g\n",
				ra2[i]*L2,ca0[i]/L2,ca1[i]/L2,ca2[i]/L2);
			}
		}

	}
	for(i=0; i<NPOINTS-1; i++) {
		ma0[i] = (ca0[i+1] - ca0[i])/(ra2[i+1] - ra2[i]);
		ma1[i] = (ca1[i+1] - ca1[i])/(ra2[i+1] - ra2[i]);
		ma2[i] = (ca2[i+1] - ca2[i])/(ra2[i+1] - ra2[i]);
	}
	cordr2 /= L2;
	if( myid == 0 ) {
		fclose(fcorfile);
	}
}

double pmf(x)
double x;
{
	double tmp1;
	double f0;

	tmp1 = cosh(A2*x);
    f0 = (L*L)*(1./(x*x)*tanh(A2*x)-A2/(tmp1*tmp1*x)
            -2.*A3/A1*x*exp(A4*x*x)-2.*A3*A4/A1*x*x*x*exp(A4*x*x)
              +A5/A1*(1.+A6*x*x+A7*x*x*x*x)*exp(A8*x*x));

	return f0;
}

double qpmf(x)
double x;
{
	double tmp1;
	double f0;

	tmp1 = cosh(A2*x);
	f0 = (L*L)*x/pow(x*x + EPSSQ*L2,1.5)
     - ((L*L)*(1./(x*x)*tanh(A2*x)-A2/(tmp1*tmp1*x)
            -2.*A3/A1*x*exp(A4*x*x)-2.*A3*A4/A1*x*x*x*exp(A4*x*x)
              +A5/A1*(1.+A6*x*x+A7*x*x*x*x)*exp(A8*x*x)));

	return f0;
}
	
GroupDirectSumInSphere(ng, glist, ParticleBin,rsphere2)
int ng;
particleStruct **glist;
particleStruct *ParticleBin;
float rsphere2;
{
	int i, monocount, quadcount;
	float epssq, x, y, z, x2, y2, z2, dr2, sr, sr2, phii;
	float fr_over_r;
	float phisum;
	particleStruct *p0;

	for(i=0; i<ng; i++) {
		p0 = glist[i];
		pos0 = p0->r;
		acc0.x = acc0.y = acc0.z = phisum = quadcount = monocount = 0;
		for(p=ParticleBin; p != NULL; p=p->nextinbin ) {

			x = (p->r).x - pos0.x;
			if( x > 0.5 ) { x -= 1.0; } else if( x < -0.5 ) { x += 1.0; };
			dr2 = x*x;
			if( dr2 > rsphere2 ) continue;

			y = (p->r).y - pos0.y;
			if( y > 0.5*yboxsize ) { y -= yboxsize; } 
			else if( y < -0.5*yboxsize ) { y += yboxsize; };

			dr2 += (y*y);
			if( dr2 > rsphere2 ) continue;

/*NOTE don't need to wrap z around */
			z = (p->r).z - pos0.z;
			dr2 += (z*z);
			if( dr2 > rsphere2 ) continue;

			CORRECTION(dr2);

			epssq = p->epssq;
			sr2 = 1.0/(dr2 + epssq);
			sr = sqrt(sr2);
			phii = (p->mass) * sr;
			phisum -= phii;
/* Note - newtonian force is just phii*sr - we are taking the difference here*/
			fr_over_r = ((phii*sr) - (p->mass)*pmforce)*sr;
			acc0.x += fr_over_r*x;
			acc0.y += fr_over_r*y;
			acc0.z += fr_over_r*z;

			monocount++;
		}
		p0->a.x += acc0.x;
		p0->a.y += acc0.y;
		p0->a.z += acc0.z;
		p0->phi += phisum;
		p0->mono += monocount;
	}
}

pforce(pcount,ng,glist)
int pcount, ng;
particleStruct **glist;
{
	int i, j, monocount, quadcount;
	float epssq, x, y, z, x2, y2, z2, dr2, sr, sr2, phii;
	float fr_over_r;
	float phisum;
	cellStruct *c;
	particleStruct *p0;
	float rsphere, rsphere2;

	rsphere = particle_radius/L;
	rsphere2 = rsphere*rsphere;

	for(i=0; i<ng; i++) {
		p0 = glist[i];
		pos0 = p0->r;
		acc0.x = acc0.y = acc0.z = phisum = quadcount = monocount = 0;
		for(j=0, c=plist; c<plist+(size_t)pcount; j++, c++) {
/*
			x = (c->r).x - pos0.x;
			y = (c->r).y - pos0.y;
			z = (c->r).z - pos0.z;

			if( x > 0.5 ) { x -= 1.0; } else if( x < -0.5 ) { x += 1.0; };
			if( y > 0.5 ) { y -= 1.0; } else if( y < -0.5 ) { y += 1.0; };
			if( z > 0.5 ) { z -= 1.0; } else if( z < -0.5 ) { z += 1.0; };

			x2 = x*x; y2 = y*y; z2 = z*z;
			dr2 = x2 + y2 + z2;
			if( dr2 > rsphere2 ) continue;
*/

			x = (c->r).x - pos0.x;
			if( x > 0.5 ) { x -= 1.0; } else if( x < -0.5 ) { x += 1.0; };
			dr2 = x*x;
			if( dr2 > rsphere2 ) continue;

			y = (c->r).y - pos0.y;
			if( y > 0.5*yboxsize ) { y -= yboxsize; } 
			else if( y < -0.5*yboxsize ) { y += yboxsize; };
			/*
			if( y > 0.5 ) { y -= 1.0; } else if( y < -0.5 ) { y += 1.0; };
			*/
			dr2 += (y*y);
			if( dr2 > rsphere2 ) continue;

/*NOTE don't need to wrap z around */
			z = (c->r).z - pos0.z;
			dr2 += (z*z);
			if( dr2 > rsphere2 ) continue;


			CORRECTION(dr2);

			epssq = c->epssq;
			sr2 = 1.0/(dr2 + epssq);
			sr = sqrt(sr2);
			phii = (c->mass) * sr;
			phisum -= phii;
/* Note - newtoninan force is just phii*sr - we are taking the difference here*/
			fr_over_r = ((phii*sr) - (c->mass)*pmforce)*sr;
			acc0.x += fr_over_r*x;
			acc0.y += fr_over_r*y;
			acc0.z += fr_over_r*z;

			monocount++;
		}
		p0->a.x += acc0.x;
		p0->a.y += acc0.y;
		p0->a.z += acc0.z;
		p0->phi += phisum;
		p0->mono += monocount;
	}
}


cforce(ccount,ng,glist)
int ccount, ng;
particleStruct **glist;
{
	int i, j, monocount, quadcount;
	float epssq, x, y, z, x2, y2, z2, dr2, sr, sr2, phii, mor3;
	float fr_over_r;
	float phisum;
	float or5, phiquad;
	float temp0, temp1, mxx, myy, mzz;
	cellStruct *c;
	particleStruct *p0;

	for(i=0; i<ng; i++) {
		p0 = glist[i];
		pos0 = p0->r;
		acc0.x = acc0.y = acc0.z = phisum = quadcount = monocount = 0;
		for(j=0, c=clist; c<clist+(size_t)ccount; j++, c++) {
			epssq = c->epssq;
			x = (c->r).x - pos0.x;
			y = (c->r).y - pos0.y;
			z = (c->r).z - pos0.z;
	
			if( x > 0.5 ) { x -= 1.0; } else if( x < -0.5 ) { x += 1.0; };
			if( y > 0.5*yboxsize ) { y -= yboxsize; } 
			else if( y < -0.5*yboxsize ) { y += yboxsize; };
			if( z > 0.5*zboxsize ) { z -= zboxsize; } 
			else if( z < -0.5*zboxsize ) { z += zboxsize; };
			/*
			if( y > 0.5 ) { y -= 1.0; } else if( y < -0.5 ) { y += 1.0; };
			if( z > 0.5 ) { z -= 1.0; } else if( z < -0.5 ) { z += 1.0; };
			*/

			x2 = x*x; y2 = y*y; z2 = z*z;
			dr2 = x2 + y2 + z2;

			CORRECTION(dr2);

			sr2 = 1.0/(dr2 + epssq);
			sr = sqrt(sr2);
			phii = (c->mass) * sr;
			phisum -= phii;
/* Note - newtoninan force is just phii*sr - we are taking the difference here*/
			fr_over_r = ((phii*sr) - (c->mass)*pmforce)*sr;
			acc0.x += fr_over_r*x;
			acc0.y += fr_over_r*y;
			acc0.z += fr_over_r*z;

#ifndef MONO

			or5 = (sr2 * sr2 * sr);
												
	/* Calculate the quadrupole potential. and sum on*/
											
#ifdef OLD
		phiquad = ( 0.5 * (qxx0 * x2 + qyy0 * y2 + qzz0 * z2 
			- epssq*momnode) + x*(qxy0*y + qxz0*z) + qyz0 * y*z ) * or5;
		phisum -= phiquad;  		
				/* Correct potential for quadrupole terms */ 
	/* Calculate the three components of the quadrupole accel. and sum on */	
		temp0 = 5.0*phiquad*sr2;											

		acc0.x += (temp0*x - (qxx0*x + qxy0*y + qxz0*z)*or5);
		acc0.y += (temp0*y - (qxy0*x + qyy0*y + qyz0*z)*or5);			
		acc0.z += (temp0*z - (qxz0*x + qyz0*y + qzz0*z)*or5);			
#endif

/* Quadrupole corrections for analytic PM force */

			mxx = qxx0 + momnode;
			myy = qyy0 + momnode;
			mzz = qzz0 + momnode;
	
			QUADCORRECTION(dr2);
	
			temp0 = 0.5*pmf1*(sr2*sr);
			temp1 = (0.5*(mxx*x2 + myy*y2 + mzz*z2) 
						+ (x*(mxy*y + mxz*z) + myz*y*z))*pmf2*or5;
	
			acc0.x += onethird*(temp1*x 
				+ temp0*((3*mxx + myy + mzz)*x + 2*(mxy*y + mxz*z)));
			acc0.y += onethird*(temp1*y 
				+ temp0*((3*myy + mzz + mxx)*y + 2*(myz*z + myx*x)));
			acc0.z += onethird*(temp1*z 
				+ temp0*((3*mzz + mxx + myy)*z + 2*(mzx*x + mzy*y)));

			quadcount++;
#else
			monocount++;
#endif

		}
		p0->a.x += acc0.x;
		p0->a.y += acc0.y;
		p0->a.z += acc0.z;
		p0->phi += phisum;
		p0->mono += monocount;
		p0->quad += quadcount;
	}
}

GroupForce(theRoot,tolsq)
cellStruct *theRoot;
float tolsq;
{
	int i, j, ng, count; 
	particleStruct *p, *glist[GROUPNUM];

	count = 0;
	for(i=0; i<NGroup; i++) {
		p=theGroupNodes[i]; j = 0;
		while( p != NULL ) {
			glist[j] = p;
			p = (particleStruct *) p->next;
			j++;
		}
		ng = j;
		count += ng;

		if( ng > 0 )
			GroupForceWalk(ng, glist, theRoot,tolsq);
	}
}

GroupForceWalk(ng,glist,theRoot,tolsq)
int ng;
particleStruct **glist;
nodeStruct *theRoot;
float tolsq;
{
	int i, pcount, ccount;
	int OpeningDecision;
	float dx1, dx2, dy1, dy2, dz1, dz2;
	cellStruct GroupSize();
	nodeStruct *nextNode;
	particleStruct *p;
	cellStruct *c;
	coordStruct pos0;

	theGroupCell = GroupSize(ng,glist);

/* Correct for self-potential of particles */
	for(i=0; i<ng; i++) {
		p = glist[i];
		p->phi = (p->mass)/sqrt(p->epssq);
	}

	ccount = 0;
	pcount = 0;
	nextNode = theRoot;
	while( nextNode != NULL ) {
		switch( nextNode->type ) {
			case PARTICULE_TYPE:
			p = (particleStruct *) nextNode;
			c = (cellStruct *) nextNode;
			COPYPARTICLE();
			pcount++;
			if( pcount == LIST_SIZE ) {
				pforce(pcount,ng,glist);
				pcount = 0;
			}
			nextNode = p->sibling;
			break;

			case CELL_TYPE:
			c = (cellStruct *) nextNode;

			if( c->nbodies <= NTREEMIN )
				OpeningDecision = 1;
			else
				OPENINGCRITERION();

			if( OpeningDecision ) { /* Can you use the cell? */
				nextNode = c->next;   /* no so go down a level */
			}
			else {
				COPYCELL();
				ccount++;
				if( ccount == LIST_SIZE ) {
					cforce(ccount,ng,glist);
					ccount = 0;
				}
				nextNode = c->sibling;
                  /* then skip to the sibling of this node */
			}
            break;

			default:
				fprintf(stderr,"illegal node type - exiting\n");
				fprintf(stderr,"node->type=%d\n",nextNode->type);
				fflush(stderr);
				EXIT(0);
			break;
		}
	}
	if( ccount > 0 ) cforce(ccount,ng,glist);
	if( pcount > 0 ) pforce(pcount,ng,glist);
}

cellStruct GroupSize(ng,glist)
int ng;
particleStruct **glist;
{
	int i;
	coordStruct rp, rmin, rmax, corner, length;
	cellStruct cell;

	rmin = glist[0]->r; rmax = glist[0]->r;
	for(i=0; i<ng; i++) {
		rp = glist[i]->r;
		if( rp.x < rmin.x ) rmin.x = rp.x;
		if( rp.y < rmin.y ) rmin.y = rp.y;
		if( rp.z < rmin.z ) rmin.z = rp.z;
		if( rp.x > rmax.x ) rmax.x = rp.x;
		if( rp.y > rmax.y ) rmax.y = rp.y;
		if( rp.z > rmax.z ) rmax.z = rp.z;
	}
	corner = rmin;
	length.x = rmax.x - rmin.x;
	length.y = rmax.y - rmin.y;
	length.z = rmax.z - rmin.z;

	cell.type   = CELL_TYPE;
	cell.corner = corner;
	cell.length = length;

	return cell;
}

GroupDirectForce(particlebin,rsphere2)
particleStruct *particlebin;
float rsphere2;
{
	int i, j, ng, count; 
	particleStruct *p, *glist[GROUPNUM];
	count = 0;

	for(i=0; i<NGroup; i++) {
		p=theGroupNodes[i]; j = 0;
		while( p != NULL ) {
			glist[j] = p;
			p = (particleStruct *) p->next;
			j++;
		}
		ng = j;
		count += ng;

		if( ng > 0 )
			GroupDirectSumInSphere(ng, glist,particlebin,rsphere2);
	}
}

ZeroGroups()
{
	int i, count;
	particleStruct *p;

/* Zero various things for each particle */
	count = 0;
	for(i=0; i<NGroup; i++) {
		p = theGroupNodes[i];
		while( p != NULL ) {
			p->a.x = p->a.y = p->a.z = p->phi = 0.0;
			p->mono = p->quad = 0;
			p->work = 0.0;
			p = (particleStruct *) p->next;
			count++;
		}
	}
}
