#include "defs.h"

int myid, numprocs;
MPI_Status receive_status;

INT8 NStackMax;
int NParticles;
INT8 NStack;
INT8 Nmax;
int Nperproc, Ntotround;
treeparticletype *TreeParticles, *theStack;
treeparticletype particleBlock[P_BLOCK_SIZE];


SendforSort(hyperLevel)
int hyperLevel;  /* Dimension of the hypercube */
{
	int i, j, count;
	int sendprocid, recvprocid;
	int sideflag;
	int master, istart, iend;
	int nsubproc;
	int dummy;
	
	nsubproc = hyperLevel >> 1;

	sendprocid = myid;
	recvprocid = sendprocid ^ (hyperLevel >> 1);  
	sideflag = (sendprocid < recvprocid);

	master = (sendprocid/nsubproc)*nsubproc;
	istart = master*Nperproc;
	iend = istart + nsubproc*Nperproc;
		/* holds the essence of the hypercube! */

/* First stack particles in our list */
	count = 0;
	for(i=0; i<NParticles; i++) {
		int id;

		id = TreeParticles[i].indx;
		if( id >= istart && id < iend ) {
			theStack[NStack] = TreeParticles[i];
			NStack++;
			if( NStack == NStackMax ) {
				fprintf(stderr,
					"Proc %d exceeding maximum allowed number in the stack %d\n", myid, NStackMax);
				/*
				ExpandParticleMemory();
				*/
			}
		}
		else {
			particleBlock[count] = TreeParticles[i];
			count++;
			if( count == P_BLOCK_SIZE ) {
				SEND(recvprocid, NP_TAG, &count, sizeof(int));
				SEND(recvprocid, PARTICLE_TAG, particleBlock,
					count*sizeof(treeparticletype));
						/* send other particles across the border */
				count = 0;
			}
		}
	}
/* Signal the receiving processor that you are finished sending */
	if( count > 0 ) {
		SEND(recvprocid, NP_TAG, &count, sizeof(int));
		SEND(recvprocid, PARTICLE_TAG, particleBlock,
					count*sizeof(treeparticletype));
	}
	SEND(recvprocid, STOP, &dummy, sizeof(int) );
}

ReceiveforSort(hyperLevel)
int hyperLevel;
{
	int i, count;
	int recvprocid, sendprocid;
	treeparticletype p;

	recvprocid = myid;
	sendprocid = recvprocid ^ (hyperLevel >> 1);  

/* continue to receive particles until you receive the stop message */
	RECEIVE(sendprocid, ANY_TAG, &count, sizeof(int));
	while( receive_status.MPI_TAG != STOP ) { 
		/* install particles on stack */
		RECEIVE(sendprocid, ANY_TAG, particleBlock, 
				count*sizeof(treeparticletype));
		for(i=0; i<count; i++) {
			theStack[NStack] = particleBlock[i];
			NStack++;
			if( NStack == NStackMax ) {
				fprintf(stderr,
				"Proc %d exceeding maximum allowed number in the stack %d\n", 
				myid, NStackMax);
				/*
				ExpandParticleMemory();
				*/
			}
		}
		RECEIVE(sendprocid, ANY_TAG, &count, sizeof(int));
	}
}

HyperSort(Np,Bp)
int *Np;
treeparticletype *Bp;
{
	int cmpare();
	int i, NProc, remainder;
	int self;
	int sendprocid, recvprocid;
	int hyperLevel;
	void hpsort();
	int imin, imax;
	int NTotal;

    MPI_Comm_size(MPI_COMM_WORLD,&numprocs);
    MPI_Comm_rank(MPI_COMM_WORLD,&myid);

	NParticles = *Np;  TreeParticles = Bp;
	Nmax = NParticles + freespace()/sizeof(treeparticletype) - 32L;
	TreeParticles = (treeparticletype *) 
		resizelast(TreeParticles,Nmax*sizeof(treeparticletype));

	theStack = TreeParticles + NParticles;
	NStackMax = Nmax - NParticles;
	NStack = 0;

/* First sort the current block of particles */
	NProc = numprocs;
	self = myid;

/* Find the sorted number in each processor and their starting id number */
	MPI_Allreduce(&NParticles,&NTotal,1,MPI_INT,MPI_SUM,MPI_COMM_WORLD);
	
	Nperproc = NTotal/NProc + 1;
	Ntotround = Nperproc*NProc;

/* Foreach level in the hypercube begin break down */
	NStack = 0;
	for(hyperLevel=NProc; hyperLevel>1; hyperLevel=hyperLevel>>1) {

		sendprocid = self;
		recvprocid = sendprocid ^ (hyperLevel >> 1); 
			/* hypercube address protocol */ 

/* condition is necessary to maintain synchronization */
/* send and receive particles from other side of domain */

		if( sendprocid < recvprocid ) {
			SendforSort(hyperLevel);
			ReceiveforSort(hyperLevel);
		}
		else {
			ReceiveforSort(hyperLevel);
			SendforSort(hyperLevel);
		}
#ifdef DEBUG
		fprintf(stderr,"hypersort:processor %d  NStack = %d\n", myid, NStack);
#endif
/* Reset the current particles */
		NParticles = NStack;
		for(i=0; i<NStack; i++)
			TreeParticles[i] = theStack[i];
		NStack = 0; /* rezero */
		theStack = TreeParticles + NParticles;
		NStackMax = Nmax - NParticles;

		SYNCHRONIZE(); /* resynchronize the processors */

	}
/* Check min and max for each processor */
	imin = imax = TreeParticles[0].indx;
	for(i=0; i<NParticles; i++) {
		if( imin > TreeParticles[i].indx ) imin = TreeParticles[i].indx;
		if( imax < TreeParticles[i].indx ) imax = TreeParticles[i].indx;
	}
	fprintf(stderr,"P%d imin %d imax %d Np %d\n",self,imin,imax,NParticles);

/* Sort the particles in place */
	hpsort(NParticles,TreeParticles-1);

	TreeParticles = (treeparticletype *) 
		resizelast(TreeParticles,NParticles*sizeof(treeparticletype));
	*Np = NParticles;
}

int cmpare(a,b)
treeparticletype *a, *b;
{
	if( a->indx > b->indx) return 1;
	if( a->indx < b->indx) return -1;
	return 0;
}

void hpsort(n,ra)
int n;
treeparticletype *ra;
{
    int i,ir,j,l;
    treeparticletype rra;

    if (n < 2) return;
    l=(n >> 1)+1;
    ir=n;
    for (;;) {
        if (l > 1) {
            rra=ra[--l];
        } else {
            rra=ra[ir];
            ra[ir]=ra[1];
            if (--ir == 1) {
                ra[1]=rra;
                break;
            }
        }
        i=l;
        j=l+l;
        while (j <= ir) {
            if (j < ir && ra[j].indx < ra[j+1].indx) j++;
            if (rra.indx < ra[j].indx) {
                ra[i]=ra[j];
                i=j;
                j <<= 1;
            } else j=ir+1;
        }
        ra[i]=rra;
    }
}
