#include "defs.h"
#include "extern_simparam.h"
#include "pm_extern.h"

#define PASSFLAG 33
#define COUNTFLAG 34
#define PARTICLEFLAG 35

typedef struct {
	float m, x, y, z, vx, vy, vz;
} phase;


basicparticletype bpblock[BLOCKSIZE];
phase pblock[BLOCKSIZE];

/*
int jwrite(ng,ngz,nspace,iseed,size,omep,omeplam, amax,astep,
			a, redshift,particles,np,myrank, nrank,nred10)
int ng, ngz, nspace, iseed;
float size, omep, omeplam, amax, astep, a, redshift;
basicparticletype *particles;
int np, myrank, nrank, nred10;
*/

int jwrite(stepcount)
int stepcount;
{
	FILE *fp;
	int i,j,k, p;
	int nptot;
	int count;
	int ncount, nblock, nremain, ncheck;
	int dummy=0;
	float m;
	char *fileheader = FILEHEADER;
	char procname[3],number[3];
	char filename[80];
	MPI_Status status;

	MPI_Allreduce(&np,&nptot, 1, MPI_INT, MPI_SUM, MPI_COMM_WORLD);
fprintf(stderr,"P%d np %d nptot %d\n",myrank,np,nptot);
#if !defined(MULTIMASS)
	m = 1.0/nptot;
#endif

	if( myrank == 0 ) {
		fprintf(stderr," writing data at a=%g,z=%g \n",a,amax/a - 1.0);
		sprintf(filename,"%s.%.5d",rvprefix,stepcount);
		fprintf(stderr," writing file: %s\n",filename);
		fp = fopen(filename,"w");
	
		ncheck = np;
		for(p=0; p<nrank; p++) {
			if( p == 0 ) {
				count = 0;
				for(k=0; k<np; k++) {
#if !defined(MULTIMASS)
					pblock[count].m = m;
#else
					pblock[count].m = particles[k].mass;
#endif
					pblock[count].x = particles[k].x;
					pblock[count].y = particles[k].y;
					pblock[count].z = particles[k].z;
					pblock[count].vx = particles[k].vx;
					pblock[count].vy = particles[k].vy;
					pblock[count].vz = particles[k].vz;
					count++;
					if( count == BLOCKSIZE ) {
						fwrite(pblock, count*sizeof(phase), 1, fp);
						count = 0;
					}
				}
				if( count > 0 ) {
					fwrite(pblock, count*sizeof(phase), 1, fp);
					count = 0;
				}
				MPI_Send(&dummy, 1, MPI_INT, myrank+1, PASSFLAG, MPI_COMM_WORLD);
			}
			else {
				do {
					MPI_Recv(&ncount, 1, MPI_INT, p, 
						COUNTFLAG, MPI_COMM_WORLD, &status);
					ncheck += ncount;
					if( ncount > 0 ) {
						MPI_Recv(pblock, ncount*sizeof(phase), MPI_BYTE, p, 
							PARTICLEFLAG, MPI_COMM_WORLD, &status);
						fwrite(pblock, ncount*sizeof(phase), 1, fp);
					}
				} while ( ncount > 0 );
			}
			fprintf(stderr,"p%d ncheck %d nptot %d\n",p,ncheck,nptot);
		}
		fclose(fp);
	}
	else {
/* Wait until you have the go ahead */
		MPI_Recv(&dummy, 1, MPI_INT, myrank-1, PASSFLAG, MPI_COMM_WORLD,&status);
		fprintf(stderr,"P%d receiving goahead...\n",myrank);
		count = 0;
		for(k=0; k<np; k++) {
#if !defined(MULTIMASS)
			pblock[count].m = m;
#else
			pblock[count].m = particles[k].mass;
#endif
			pblock[count].x = particles[k].x;
			pblock[count].y = particles[k].y;
			pblock[count].z = particles[k].z;
			pblock[count].vx = particles[k].vx;
			pblock[count].vy = particles[k].vy;
			pblock[count].vz = particles[k].vz;
			count++;
			if( count == BLOCKSIZE ) {
				MPI_Send(&count, 1, MPI_INT, 0, COUNTFLAG, MPI_COMM_WORLD);
				MPI_Send(pblock, count*sizeof(phase), MPI_BYTE, 0, 
					PARTICLEFLAG, MPI_COMM_WORLD);
				count = 0;
			}
		}
		if( count > 0 ) {
			MPI_Send(&count, 1, MPI_INT, 0, COUNTFLAG, MPI_COMM_WORLD);
			MPI_Send(pblock, count*sizeof(phase), MPI_BYTE, 0, 
				PARTICLEFLAG, MPI_COMM_WORLD);
			count = 0;
		}
		count = 0;
/* Signal that you are done */
		MPI_Send(&count, 1, MPI_INT, 0, COUNTFLAG, MPI_COMM_WORLD);
/* Tell the next guy to start transferring data */
		if( myrank < nrank - 1 )
			MPI_Send(&dummy, 1, MPI_INT, myrank+1, PASSFLAG, MPI_COMM_WORLD);
	}

	if( myrank == 0 ) {
		FILE *paramfile;
		char paramname[80];

		sprintf(paramname,"params.%.5d",stepcount);
		paramfile = fopen(paramname,"w");
		
		fprintf(paramfile,"%.8g %.8g\n",boxsize_hinv_mpc,hubble_parameter);
	    fprintf(paramfile,"%d %.8g %.8g %.8g %.8g\n",
           npow0,omega,lambda,bias_8mpc,rsmooth);
		fprintf(paramfile,"%d %d\n",nmesh, nspace0);
		fprintf(paramfile,"%.8g %.8g %.8g\n",
				SPHERE_RADIUS,particle_radius,theta);
		fprintf(paramfile,"%.8g %.8g %.8g\n",z_init,delta_a,a);
		fprintf(paramfile,"%d %d %d\n",nstep,stepcount,nskip);
		fprintf(paramfile,"%d\n",seed);
		fprintf(paramfile,"%s %s\n",filename, rvprefix);
		fprintf(paramfile,"%.8g %.8g %.8g %.8g\n",
			ken, ktot, const0, poten0);
		fprintf(paramfile,"%s\n",camerafilename);

		fclose(paramfile);
	}
	return 0;
}

/*
void jread(int *ng,int *ngz,int *nspace,int *iseed,float *size,float *omep,
		float *omeplam, float *amax,float *astep,float *pconv,
		float *vconv,float *a, float *redshift,basicparticletype *particles, 
		int *np,int myrank, int nrank,int *nstep)
*/
void jread()
{
	int i, j, k, nmax, count, indx;
	int nsend, nrecv;
	int passflag;
	long NTotal, NLocal, NLocal0, NTotalcheck;
	long nbytes;
	float m, mrv[7];
	int ptrsize();
	FILE *rvfile;
	char rvname[80];
	MPI_Status status;
	float ksum, totalken;
	basicparticletype *p;

/* passed the test already for the existence of this file */
	if( myrank == 0 ) {
		rvfile = fopen(rvfilename,"r");
		fseek(rvfile,0,2); nbytes = ftell(rvfile); fseek(rvfile,0,0);
		NTotal = nbytes/(7*sizeof(FLOAT_TYPE));
		fprintf(stderr,"In file: NTotal %d\n",NTotal);
	}
	MPI_Bcast(&NTotal, 1, MPI_LONG, 0, MPI_COMM_WORLD);
	fprintf(stderr,"P%d has NTotal %d\n",myrank,NTotal);
	NLocal = NTotal/nrank;
	if( myrank == 0 ) {
		NLocal0 = NLocal;
		NLocal += (NTotal - NLocal*nrank);
	}
	MPI_Allreduce(&NLocal,&NTotalcheck, 1, MPI_LONG, MPI_SUM, MPI_COMM_WORLD);
	fprintf(stderr,"Checking NTotal %d = NTotalcheck %d\n",NTotal, NTotalcheck);
	fprintf(stderr,"P%d NLocal %d\n",myrank,NLocal);
	np = NLocal;

#if !defined(MULTIMASS)
	m =  1.0/NTotal;
#endif
	
	nmax = ptrsize(particles)/sizeof(basicparticletype);
	fprintf(stderr,"P%d nmax %d\n",myrank,nmax);

	if( myrank == 0 ) {
		indx = 0;
		for(i=0; i<NLocal; i++) {
			float mrv[7];

			fread(mrv,28,1,rvfile);
#if defined (MULTIMASS)
			particles[i].mass = mrv[0];
#endif
			particles[i].x = mrv[1];
			particles[i].y = mrv[2];
			particles[i].z = mrv[3];
			particles[i].vx = mrv[4];
			particles[i].vy = mrv[5];
			particles[i].vz = mrv[6];
			particles[i].freework = 1.0;
			particles[i].indx = indx;
			indx++;
		}
		for(j=1; j<nrank; j++) {
			count = 0;
			nsend = NLocal0/BLOCKSIZE;
			if( (NLocal0 - nsend*BLOCKSIZE) > 0 ) nsend++;
			MPI_Send(&nsend, 1, MPI_INT, j, GOFLAG, MPI_COMM_WORLD);
fprintf(stderr,"sending %d blocks to P%d\n",nsend,j);
			for(i=0; i<NLocal0; i++) {
				fread(mrv,28,1,rvfile);
#if defined(MULTIMASS)
				bpblock[count].mass = mrv[0];
#endif
				bpblock[count].x = mrv[1];
				bpblock[count].y = mrv[2];
				bpblock[count].z = mrv[3];
				bpblock[count].vx = mrv[4];
				bpblock[count].vy = mrv[5];
				bpblock[count].vz = mrv[6];
				bpblock[count].freework = 1.0;
				bpblock[count].indx = indx;
				count++;
				indx++;
				if( count == BLOCKSIZE ) {
					MPI_Send(bpblock, count*sizeof(basicparticletype),
						MPI_BYTE, j, PASSFLAG, MPI_COMM_WORLD);
					count = 0;
				}
			}
			if( count > 0 ) {
				MPI_Send(bpblock, count*sizeof(basicparticletype),
					MPI_BYTE, j, PASSFLAG, MPI_COMM_WORLD);
				count = 0;
			}
		}
		fclose(rvfile);
	}
	else {
		MPI_Recv(&nrecv, 1, MPI_INT, 0, GOFLAG, MPI_COMM_WORLD, &status);
fprintf(stderr,"%d receiving %d blocks\n",myrank,nrecv);
		k = 0;
		for(i=0; i<nrecv-1; i++) {
			MPI_Recv(particles+k,BLOCKSIZE*sizeof(basicparticletype),
				MPI_BYTE, 0, PASSFLAG, MPI_COMM_WORLD, &status);
		fprintf(stderr,"P%d received %d particles\n",myrank,BLOCKSIZE);
			k += BLOCKSIZE;
		}
		count = NLocal - (nrecv-1)*BLOCKSIZE;
		MPI_Recv(particles+k,count*sizeof(basicparticletype),
			MPI_BYTE, 0, PASSFLAG, MPI_COMM_WORLD, &status);
		fprintf(stderr,"P%d received %d particles\n",myrank,count);
	}

/* Initialize kinetic energy term */
	ksum = 0;
	for(p=particles; p<particles+NLocal; p++) {
		ksum += (p->mass)*((p->vx)*(p->vx) + (p->vy)*(p->vy) + (p->vz)*(p->vz));
	}
	MPI_Allreduce(&ksum,&totalken,1,MPI_FLOAT,MPI_SUM,MPI_COMM_WORLD);
	ken = totalken;
	np = NLocal;
}
