/* My GOTPM code using the tree corrections developed by Juhan Kim
 * =======================================================================
 * It is in an interim status to the 4096^3 particle simulations at PSC.
 *-------------------------------------------------------------------------
 * Comparison to the original code has been checked with a success. 
 * 		Jan. 17th 2006
 * Time evolution has been checked ? Jan. 18th 2006
 * FoF group test has been passed ? Jan. 19th 2006
 ***************************************************************************
 * In mytreecorrection, there is no preprocessor for the sake of simplicity.
 */
#include<stdio.h>
#include<math.h>
#include<mpi.h>
#include<stddef.h>
#include <stdlib.h>
#include<sys/times.h>
#include "pmheader.h"
#include "common_simparam.h"
#include "pm_common.h"
#include "Memory.h"

#define INFINITY -1
#define max(A,B) ( (A) > (B) ? (A) : (B) )
#define min(A,B) ( (A) < (B) ? (A) : (B) )

void pm(int, int, int, int, int, int);

void MAIN_(int argc, char *argv[])
{
	int nx, ny, nz, local_nz;
	int nspace;
	int ierror=1, iwrite=0, icont;
	int nstep = 0;
    char c;
	int memflag, InitializeBigMemory();
	FILE *simfile;

	MPI_Init(&argc,&argv);
	MPI_Comm_rank(MPI_COMM_WORLD,&myrank);
	MPI_Comm_size(MPI_COMM_WORLD,&nrank);

	if( argc == 3 ) {
		if( !(simfile = fopen(argv[1],"r")) ) {
		    fprintf(stderr,"Can't open %s - exiting\n",argv[1]);
			MPI_Finalize();
			exit(0);
		}
		maxtime = atof(argv[2]);
   	}
	else {
		fprintf(stderr,"usage: pyul paramsfile maximum_run_time\n");
		MPI_Finalize();
		exit(0);
	}

	ReadSimulationParameters(simfile,&icont);

	/*
	nx = ny = nz = nmesh;
	*/
	nx = nxmesh; ny = nymesh; nz = nzmesh;
	nspace = nspace0;
	rth = rtophat_mesh;
	local_nz = nz/nrank;
	np = nx * ny * local_nz / nspace / nspace / nspace ;
    densize = 2*(nx/2+1) * ny * (local_nz+8);

	/* if using grid force 4 or 1*/
#ifdef ENOUGH_SPACE
	worksize = max(densize,(nx+4)*(ny+4)*(local_nz+4));
#else
  	worksize =  densize;
#endif
  	den = work;

/* Initialize re-useable memory array and assign it to particles */
/*
	memflag = InitializeBigMemory();
	if( !memflag ) {
		fprintf(stderr,"Can't initialize memory - aborting job\n");
		MPI_Finalize();
		exit(0);
	}
*/
	if(Make_Total_Memory() == 0 ){
		fprintf(stderr,"Can't initialize memory - aborting job\n");
		MPI_Finalize();
		exit(0);
	}
	
	/*
	mallocptr = (void **)(&particles);
	particles = (basicparticletype *) Malloc(np*sizeof(basicparticletype),
			mallocptr);
			*/
	particles = (basicparticletype *) Malloc(np*sizeof(basicparticletype),
			PPTR(particles));

	pk = (float *)calloc(nx+1,sizeof(float));
	npk = (float *)calloc(nx+1,sizeof(float));
	cor = (float *)calloc(nx+1,sizeof(float));
	tpk = (float *)calloc(nx+1,sizeof(float));
	tnpk = (float *)calloc(nx+1,sizeof(float));
	tcor = (float *)calloc(nx+1,sizeof(float));
	corw = (float *)calloc(nx,sizeof(float));
	tcorw = (float *)calloc(nx,sizeof(float));
	xabp0 = (int *)calloc(nx+21,sizeof(int));
	xabp1 = (int *)calloc(nx+21,sizeof(int));
	xabm1 = (int *)calloc(nx+21,sizeof(int));
	yabp0 = (int *)calloc(ny+21,sizeof(int));
	yabp1 = (int *)calloc(ny+21,sizeof(int));
	yabm1 = (int *)calloc(ny+21,sizeof(int));
#ifdef VarPM
	zabp0 = (int *)calloc(nz+28,sizeof(int));
	zabp1 = (int *)calloc(nz+28,sizeof(int));
	zabm1 = (int *)calloc(nz+28,sizeof(int));
#else
	zabp0 = (int *)calloc(local_nz+28,sizeof(int));
	zabp1 = (int *)calloc(local_nz+28,sizeof(int));
	zabm1 = (int *)calloc(local_nz+28,sizeof(int));
#endif
	rcounts = (int *)calloc(nrank,sizeof(int));
	displs = (int *)calloc(nrank,sizeof(int));

	fprintf(stderr,"icont %d\n",icont);
	iwrite = 0;
	if( icont == 0 ) { iwrite = 1; }
    pm(icont,iwrite, nx, ny, nz, local_nz);

	MPI_Barrier(MPI_COMM_WORLD);
	MPI_Finalize();
	exit(0);
}

int ReadSimulationParameters(simfile,icont)
FILE *simfile;
int *icont;
{
        char filename[180];
	FILE *rvfile;
	int abortflag;

    /* Box size */
	fscanf(simfile,"%f %f\n",&boxsize_hinv_mpc,&hubble_parameter);

	/* Power Spectrum */
	fscanf(simfile,"%f %f %f %f %f %f\n",
           &npow0,&omega,&omegabaryon, &lambda,&bias_8mpc,&rsmooth);
 
	/* Mesh Parameters */
	fscanf(simfile,"%d %d %d %d\n",&nxmesh, &nymesh, &nzmesh,&nspace0);

    /* Tree parameters */
	fscanf(simfile,"%f %f %f\n",&SPHERE_RADIUS,&particle_radius,&theta);
	L = nxmesh; /* Set L parameter */
	L2 = L*L;
		 
	/* Time Parameters */
	fscanf(simfile,"%f %f %f\n",&z_init,&delta_a,&a);
	amax = 1.0 + z_init;

	/* Stepping Parameters */
	fscanf(simfile,"%d %d %d\n",&nstep,&stepnum,&nskip);
		 
	/* Negative integer for random number generator for initial conditions */
    fscanf(simfile,"%d\n",&seed);
		 
	rtophat_mesh = nxmesh/boxsize_hinv_mpc*8.0;

    /* RV file name*/
	fscanf(simfile,"%s %s\n",rvfilename, rvprefix);

    /* Energy parameters */
	fscanf(simfile,"%f %f %f %f\n",
		&ken, &ktot, &const0, &poten0);

    /* Camera file name */
	fscanf(simfile,"%s\n",camerafilename);

    fclose(simfile);

/* Ok now output to stdout to verify all is well */
	if( myrank ==0 ) {
		fprintf(stdout,"*** INPUT PARAMETER LIST ***\n\n");
		fprintf(stdout,"*** Cosmological Parameters ***\n");
		fprintf(stdout,"Box size = %g hinv_Mpc   Hubble_parameter = %g\n",
			boxsize_hinv_mpc, hubble_parameter);
		fprintf(stdout,"npow = %g Omega = %g OmegaBaryon = %g Lambda = %g\nBias_8Mpc = %g Rsmooth = %g\n\n", npow0,omega,omegabaryon,lambda,bias_8mpc,rsmooth);
		
		fprintf(stdout,"\n*** Mesh Parameters ***\n");
		fprintf(stdout,"Nxmesh = %d Nymesh = %d Nzmesh = %d Spacing = %d\n\n",
				nxmesh,nymesh,nzmesh,nspace0);
		fprintf(stdout,"\n*** Tree Parameters ***\n");
		fprintf(stdout,"SPHERE_RADIUS = %g Theta = %g\n\n",SPHERE_RADIUS,theta);
		fprintf(stdout,"\n*** Time Parameters ***\n");
		fprintf(stdout,"Initial Redshift = %g Delta_a = %g Current a = %g\n\n",z_init,delta_a,a);
		fprintf(stdout,"\n*** Stepping Parameters ***\n");
		fprintf(stdout,"Nstep = %d Stepnum = %d Nskip = %d\n", 
			nstep,stepnum,nskip);
		fprintf(stdout,"Random Seed = %d\n\n",seed);
		fprintf(stdout,"Initial RV File %s  RV File Prefix %s\n\n",
			rvfilename,rvprefix);
	}
	fflush(stdout);

/* Check to see if the initial conditions file exists - if not then default to g   generating your own initial conditions */

	abortflag = 0;
	if( myrank == 0 ) {

/* if strcmp() == 0 then string comparison is successful */
		if( strcmp(rvfilename,"INITIAL") == 0 ) {
			fprintf(stderr,"Generating Initial Conditions for run....\n");
			*icont = 0;
			a = 1.0;
		}
		else {
                        sprintf(filename,"%s%.5d",rvfilename,0);
			rvfile = fopen(filename,"r");
			if( rvfile == NULL ) {
				fprintf(stderr,"Can't open initial rvfile %s - exiting\n",	
					rvfilename);
				abortflag = 1;
			}	
			else { 
				fprintf(stderr,"Successfully opened rvfile %s - starting simulation\n",rvfilename);
				fclose(rvfile); 
				*icont = 1;
			}
		}
	}
	MPI_Bcast(&abortflag, 1, MPI_INT, 0, MPI_COMM_WORLD);
	if( abortflag ) {
		fprintf(stderr,"Aborting can't find initial rvfile\n");
		MPI_Finalize();
		exit(0);
	}

	MPI_Bcast(icont, 1, MPI_INT, 0, MPI_COMM_WORLD);
/* SAMPLE PARAMETER FILE 
 * 65 0.65
 * 1.0 0.3 0.7 1.0 0.0
 * 128 1
 * 4 1.0
 * 71 .10
 * 700 0 100
 * -1
 */
		 
}
			 
/* Initialization for power spectrum and initial condition generator */
			 
void getpowerparams_(nspace,size,amax,astep,npow,smooth,h,rth,bias,iseed,
		omepb)
int *nspace;
float *size, *amax, *astep;
float *npow;
int *iseed;
float *smooth, *h, *rth, *bias;
float *omepb;
{
 
     *nspace = nspace0;
     *size = boxsize_hinv_mpc;
     *amax = 1.0 + z_init;
     *astep = delta_a;
     *npow = npow0;
     *iseed = seed;
     *smooth = rsmooth;
     *h = hubble_parameter;
     *rth = rtophat_mesh;
     *bias = bias_8mpc;
	 *omepb = omegabaryon;
}
