#include<stdio.h>
#include<stdlib.h>
#include<string.h>
#include<math.h>
#include "/export/local/include/fitsio.h"
#include "/export/local/include/longnam.h"
#define sliceheader "slice"
#define SQR(A) ((A)*(A))

#define X0 0.
#define Y0 0.
#define Z0 0.

float omega0,H,lambda0;
float func1(float a){
    return 2997.92L/H/sqrt((1.+a)*(1.+a)*(1.+a)*omega0
	            +(1.+a)*(1.+a)*(1.-omega0-lambda0) +lambda0);
}

float qsimp(float (*func)(float),float,float);

void z2comoving(float *R, float *dR1, float *dR2,int nx, 
		float amax,float a, float astep, 
		float omep,float omeplam,float size, float h){
	float initz = 0;
	float shellwidth1,shellwidth2;
	float redshift = amax/a - 1.;
	shellwidth1 = amax/(a+astep*0.5)-1.;
	shellwidth2 = amax/(a-astep*0.5)-1.;
	if(omep == 1.0) {
		*R = 2.L*2997.92l/h * (1.-1./sqrt(1.+redshift));
		*dR1 = 2.L*2997.92L/h*(1.-1./sqrt(1.+shellwidth1));
		*dR2 = 2.L*2997.92L/h*(1.-1./sqrt(1.+shellwidth2));
	}
	else {
		omega0 = omep;
		lambda0 = omeplam;
		H = h;
		*R = qsimp(func1,initz,redshift);
		*dR1 = qsimp(func1,initz,shellwidth1);
		*dR2 = qsimp(func1,initz,shellwidth2);
	}
	*R = *R/size*nx;
	*dR1 = *dR1/size*nx;
	*dR2 = *dR2/size*nx;
	*dR1 = *R - *dR1;
	*dR2 = *dR2 - *R;
}
float redshift2dist(float z,float omep,float omeplam,float h){
	float dist;
	float initz = 0.;
	if(omep == 1.0){
		dist = 2.L*2997.92L/h * ( 1.-1./sqrt(1.+z));
		return dist;
	}
	else {
		omega0 = omep;
		lambda0 = omeplam;
		H = h;
		dist = qsimp(func1,initz,z);
		return dist;
	}
}

int main(int argc, char **argv){
	int i,j,k;
	float size,hubble;
	int npow;
	float omep,omeplam,bias,rsmooth;
	int ng,nspace;
	float Sphere_radius,particle_radius,theta;
	float zinit,astep,a;
	int  nstep,stepnum,nskip;
	int  iseed;
	
	FILE *params;
	FILE *datafile;
	char datafilename[80];
	int nx,ny,nz;
	int nxim,nyim;
	float *den;
	float *image;
	int nx0,ny0;
	int iy,ix;
	int amax;
	float z,*dist,*angle,*zptr,*distptr,*angleptr;

	if(argc != 5) {
		fprintf(stderr,"Usage: slice [params.dat] [image size nx, ny] [fitsname]\n");
		exit(0);
	}
	if((params = fopen(argv[1],"r")) == NULL) {
		fprintf(stderr,"Error opening file %s\n",argv[1]);
		exit(0);
	};
	nxim = atoi(argv[2]);
	nyim = atoi(argv[3]);
	image = (float *) malloc(sizeof(float)*nxim*nyim);
	nx0 = nxim; ny0 = nyim/2;
	fscanf(params,"%f %f\n",&size,&hubble);
	fscanf(params,"%d %f %f %f %f\n",&npow,&omep,&omeplam,&bias,&rsmooth);
	fscanf(params,"%d %d\n",&ng, &nspace);
	fscanf(params,"%f %f %f\n",&Sphere_radius,&particle_radius,&theta);
	fscanf(params,"%f %f %f\n",&zinit,&astep,&a);
	fscanf(params,"%d %d %d\n",&nstep,&stepnum,&nskip);
	fscanf(params,"%d\n",&iseed);
	amax = (zinit+1);
	for(i=0;i<nxim*nyim;i++){
		image[i] = -1.;
	}

	dist = (float *) malloc(sizeof(float)*nxim*nyim);
	angle = (float *) malloc(sizeof(float)*nxim*nyim);
	distptr = dist;angleptr=angle;
	for(iy=0;iy<nyim;iy++)
		for(ix=0;ix<nxim;ix++){
				z = sqrt(pow(ix-nx0,2) + pow(iy-ny0,2))/nxim * zinit;
				/* here multiply factor nx is omitted
				 * because We don't know the value of nx before reading
				 * slice files */
				*distptr = redshift2dist(z,omep,omeplam,hubble)/size /* *nx */;
				*angleptr = atan((float)(iy-ny0)/(float)(ix-nx0));
				distptr++;angleptr++;
		}
	for(i=1;i<=nstep;i++){
		float R,dR1,dR2;
		sprintf(datafilename,"%s.%.5d",sliceheader,i);
		datafile = fopen(datafilename,"rb");
		fread(&nx,sizeof(int),1,datafile);
		fread(&ny,sizeof(int),1,datafile);
		den = (float *) malloc(sizeof(float)*nx*ny);
		fread(den,sizeof(float),nx*ny,datafile);
		fclose(datafile);
		z2comoving(&R, &dR1, &dR2,nx, amax,a, astep, omep,omeplam,size, hubble);
		printf("Now %d's slice calculation\n",i);
		distptr = dist;angleptr=angle;
		for(iy=0;iy<nyim;iy++)
			for(ix=0;ix<nxim;ix++) {
				int jx,jy;
				int r;
				if(*distptr*nx >= R-dR1 && *distptr*nx< R+dR2){
					jx = cos(*angleptr)**distptr*nx;
					jy = sin(*angleptr)**distptr*nx;
					if(*angleptr>=0.) {
						jx = -jx;
						jy = -jy;
					}
					jx = jx + nx*10000;
					jy = jy + ny*10000;
					jx = jx%nx;
					jy = jy%ny;
					image[ix+nxim*iy] = den[jx+nx*jy];
				}
				distptr++;angleptr++;
			}
		printf("end of %d's slice calculation\n",i);
		free(den);
		a = a + astep;
	}
	{
		fitsfile *fptr;
		char card[FLEN_CARD];
		int nkeys,ii,status=0;
		int naxis = 2;
		INT8 nelements = nxim * nyim;
		INT8 naxes[2];
		INT8 fpixel[2];
		char filename[80];
		naxes[0] = nxim;
		naxes[1] = nyim;
		fpixel[0] = fpixel[1] = 1;
		sprintf(filename,"\!%s",argv[4]);
		printf("%s\n",filename);
		(void)fits_create_file(&fptr,filename,&status);
		if(status) fits_report_error(stderr,status);
		(void) fits_create_img(fptr,FLOAT_IMG,naxis,naxes,&status);
		if(status) fits_report_error(stderr,status);
		(void)fits_write_pix(fptr,TFLOAT,fpixel,nelements,
							 image,&status);
		if(status) fits_report_error(stderr,status);
		(void)fits_close_file(fptr,&status);
		if(status) fits_report_error(stderr,status);
	}
	return 0;
}
