#include "defs.h"
#include "common.h"
#include "extern_simparam.h"
extern float cputime0[100], cputime1[100];

#define TO_TREE_SCALING 1
#define TO_PM_SCALING 2
#define DDSKIP 10
#define wrap(a,b,c) { a = ((b)+(c)) % (c); }
#define MIN(A,B) ((A)<(B)? (A):(B))

extern basicparticletype *particles;
void migrate(int *,int,float,float);
int NBins;
static int nx, ny, nz, dbin;
int  NBasicParticles;
static int BinCount[NBINS];
static basicparticletype *BasicParticleBin[NBINS];
static particleStruct *ParticleBin[NBINS];
static cellStruct *BinRoot[NBINS];
static int treestepcount=0;
#ifdef VarPM
extern float zstart, zheight;
#else
static float zstart, zheight;
#endif
float xboxup,yboxup,zboxup,xboxsize,yboxsize,zboxsize;

/*
int	GetVisitors(int, basicparticletype *, float, float, float);
*/
int	GetVisitors();

float pmass0, epssq0;
double disp,tdisp;
int tnp;
#define MAX(A,B) ((A)>(B) ? (A):(B))

#if !defined (INDVEPS) && !defined(MULTIMASS)
int TreeCorrection(nobj,BasicParticles,fact1,fact2,pfact,pmass,epssq,stepcount)
#elif MULTIMASS
int TreeCorrection(nobj,BasicParticles,fact1,fact2,pfact,epssq,stepcount)
#else
int TreeCorrection(nobj,BasicParticles,fact1,fact2,pfact, stepcount)
#endif
int *nobj;
basicparticletype *BasicParticles;
float fact1, fact2, pfact;
#if !defined (MULTIMASS)
float pmass;
#endif
#if !defined (INDVEPS)
float epssq;
#endif
int stepcount;
{
	int binnum, ix, iy, iz;
	int bcount, checkcount;
	int i, j, k, i0, j0, count, icount;
	int ix0, iy0, iz0, ix1, iy1, iz1, ixw, iyw;
	int visitorflag0, visitorflag1;
	double totalwork;
	double monotot, quadtot;
	particleStruct *p;
	coordStruct refCorners[20], refLengths[20];
	coordStruct dcorner, dlength, bincorner;
	coordStruct rmin0, rmax0;
	cellStruct *BuildTree();
	float binsize, thetasq, minlengthfrac;
	float rsphere, rsphere2, prad2;
	float procspeed;
	static int startflag=1;
	int NVisitors; 
	basicparticletype *bp;
	char accname[80];
	FILE *accfile;

    MPI_Comm_size(MPI_COMM_WORLD,&numprocs);
    MPI_Comm_rank(MPI_COMM_WORLD,&myid);

TIMER_START(55);
TIMER_START(2);

	NBasicParticles = *nobj;
	minlengthfrac = 1.0;
	if(stepcount <=110) theta = 0.5;
	else theta = 0.8;
	thetasq = theta*theta;
/* WARNING - L must be divisible by SPHERE_RADIUS or BIG PROBLEMS */
	if( startflag ) {
		/*
		nx = ny = (int) floor(L/SPHERE_RADIUS); 
		ny = (int) floor((float)nymesh/SPHERE_RADIUS);
		*/
		nx = (int) floor(L/SPHERE_RADIUS);
		ny = (int) floor((float)nymesh/SPHERE_RADIUS);
		SPHERE_RADIUS = ((float) L)/nx;
		if( myid == 0 ) {
			fprintf(stderr,"New SPHERE_RADIUS = %g\n",SPHERE_RADIUS);
		}
	}
		xboxup = 0.5;
		yboxup = (float)nymesh/(float)nxmesh-0.5;
		zboxup = (float)nzmesh/(float)nxmesh-0.5;
		xboxsize = 1.;
		yboxsize = (float)nymesh/(float)nxmesh;
		zboxsize = (float)nzmesh/(float)nxmesh;
		printf("P%d has tree box size %g-%g,  %g-%g, %g-%g\n",myid,
				xboxsize,xboxup,yboxsize,yboxup,zboxsize,zboxup);
	rsphere = ((float) SPHERE_RADIUS)/L;
	binsize = rsphere;
	prad2 = (particle_radius*particle_radius)/(L*L);
#if !defined(MULTIMASS)
	pmass0 = pmass;
#endif
#ifndef INDVEPS
	epssq0 = epssq;
#endif

	if( startflag ) {
		InitCorrectionFunction();
	}
	if( startflag ) startflag = 0;

	dbin = 1;


	RescaleParticles(NBasicParticles,BasicParticles,TO_TREE_SCALING);

#ifdef DEBUG
	for(bp=BasicParticles; bp<BasicParticles+(size_t)NBasicParticles; bp++) {
		if( myid == 0 )
			if( bp->z > 0.0 ) fprintf(stderr,"P%d bp->z %g\n",myid,bp->z); 
		if( myid == numprocs-1 )
			if( bp->z < 0.0 ) fprintf(stderr,"P%d bp->z %g\n",myid,bp->z); 
	}
#endif

	TIMER_START(0);
	if( treestepcount % DDSKIP == 0 || treestepcount < 3 ) {
		DomainDecomposition(&NBasicParticles,BasicParticles,
			&dcorner,&dlength,refCorners,refLengths);
		zstart = dcorner.z;
		zheight = dlength.z;
		BasicParticles = particles;
	}
	else {
#ifndef VarPM
		migrate(&NBasicParticles, 1, zheight, zstart);
#endif
	}
	TIMER_STOP(0);
	printf("P%d has %d particles\n",myid,NBasicParticles);
	SYNCHRONIZE();
	if( myid == 0 ) 
		fprintf(stderr,"CPU(Domain decomp) %g\n",ELAPSED_TIME(0));

	MaximumExtent(NBasicParticles,BasicParticles,&rmin0,&rmax0);
	dcorner = rmin0;
	dlength.x = rmax0.x - rmin0.x;
	dlength.y = rmax0.y - rmin0.y;
	dlength.z = rmax0.z - rmin0.z;
/* Ok check to see if your particles are in their domains */
	for(bp=BasicParticles; bp<BasicParticles+(size_t)NBasicParticles; bp++) {
		if( bp->z < dcorner.z || bp->z > dcorner.z+dlength.z)
		fprintf(stderr,"P%d there's a problem in Domain Decomposition\n",myid);
	}

#ifdef DEBUG
	fprintf(stderr,"P%2d N=%6d    % 8.6f < z < % 8.6f height=% 8.6f\n",
		myid, NBasicParticles, dcorner.z, dcorner.z+dlength.z, dlength.z);
#endif

#ifdef OLDVISITOR
/* Check to see if the interaction sphere radius is less than the slice width */	
	visitorflag0 = 1;
	if( dlength.z < rsphere ) {
		fprintf(stderr,"P%d Slice too thin - use fewer processors or change GetVisitors function - slice height=%g rsphere=%g\n", myid, dlength.z, rsphere);
		visitorflag0 = 0;
	}
	MPI_Allreduce(&visitorflag0,&visitorflag1,1,MPI_INT,MPI_SUM,MPI_COMM_WORLD);

	if( visitorflag1 != numprocs ) {
		MPI_Finalize();
		exit(0);
	}
	SYNCHRONIZE();
#endif
/* Check to make sure every processor has a wide enough domain */

	TIMER_START(67);
	dcorner.x = -0.5;
	dcorner.y = -0.5;
	NVisitors = 0;
#ifndef NOVISITORS
	NVisitors = GetVisitors(NBasicParticles,BasicParticles,
		rsphere,dcorner.z,dlength.z);
	SYNCHRONIZE();
#endif
#ifdef DEBUG
	fprintf(stderr,"P%d NVisitors = %d\n",myid,NVisitors);fflush(stdout);
#endif
	MPI_Barrier(MPI_COMM_WORLD);fflush(stdout);
	TIMER_STOP(67);
	if( myid == 0 ) {
			fprintf(stderr,"CPU (GetVisitors)=%g\n",ELAPSED_TIME(67));
	}

	MaximumExtent(NBasicParticles+NVisitors,BasicParticles,&rmin0,&rmax0);
	/*
	rmin0.x=rmin0.y=rmin0.z=1.E28;
	rmax0.x=rmax0.y=rmax0.z=-1.E28;
	for(bp=BasicParticles; bp<BasicParticles+(size_t)NBasicParticles+(size_t)NVisitors; bp++) {
		rmin0.x = MIN(rmin0.x,bp->x);
		rmin0.y = MIN(rmin0.y,bp->y);
		rmin0.z = MIN(rmin0.z,bp->z);
		rmax0.x = MAX(rmax0.x,bp->x);
		rmax0.y = MAX(rmax0.y,bp->y);
		rmax0.z = MAX(rmax0.z,bp->z);
	}
	*/
	for(bp=BasicParticles; bp<BasicParticles+(size_t)NBasicParticles+(size_t)NVisitors; bp++) {
		if( bp->z < rmin0.z || bp->z > rmax0.z) {
		fprintf(stderr,"P%d there's a problem in Maximum Extent\n",myid);
		EXIT(0);
		}
	}
	bp=BasicParticles;
	printf("P%d has  %d particles %g %g\n",myid,NBasicParticles, rmin0.z, rmax0.z);

/*
	nz = (int) ceil((rmax0.z-rmin0.z)/binsize) + 1;
	ny = (int) ceil((rmax0.y-rmin0.y)/binsize) + 1;
*/
	nz = (int) ceil((rmax0.z-rmin0.z)/binsize) + 1;
	NBins = nx*ny*nz;
#ifdef DEBUG
	fprintf(stderr,"P%d %g < z < %g binsize %g nz %d\n",
			myid,rmin0.z,rmax0.z,binsize,nz);
#endif
	
	if( NBins > NBINS )
		fprintf(stderr,"not enough Bins\n");
#ifdef DEBUG
	fprintf(stderr,"P%d NBins %d\n",myid,NBins);
#endif

	BinParticles(NBasicParticles+NVisitors,BasicParticles,
		binsize,rmin0,nx,ny,nz,BasicParticleBin);
#ifdef DEBUG
	fprintf(stderr,"P%d binned particles nx ny nz NBins %d %d %d %d\n",myid,
		nx, ny, nz, NBins);
#endif

/* Count particle in bins */

	for(i=0; i<NBins; i++) {
		BinRoot[i] = NULL;
		if( BasicParticleBin[i] != NULL ) {
			count = 0;
			for(bp=BasicParticleBin[i]; bp != NULL; bp=bp->u.nextinbin) 
				count++;
			BinCount[i] = count;
		}
		else {
			BinCount[i] = 0;
		}
	}

/* An OUTER LOOP WILL START HERE */
/* Do a subdomain of trees */

	monotot = quadtot = 0.0;
	checkcount = 0;

	iz0 = 0; iz1 = nz;
	iy0 = 0; iy1 = ny;
	ix0 = 0; ix1 = nx;
	for(ix0=0; ix0<nx; ix0+=8) {
	for(iy0=0; iy0<ny; iy0+=8) {
	ix1 = ix0 + 8; iy1 = iy0 + 8;

	InitializeTreeParticles(BasicParticles,ix0-1,ix1+1,iy0-1,iy1+1,iz0,iz1);

    TIMER_START(1);

    for(iz=iz0; iz<iz1; iz++) {
    for(iy=(iy0-1); iy<(iy1+1); iy++) { wrap(iyw,iy,ny);
    for(ix=(ix0-1); ix<(ix1+1); ix++) { wrap(ixw,ix,nx);
        binnum = ixw + nx*(iyw + ny*iz);
        if( BinCount[binnum] > NTREEMIN ) {
            bincorner.x = rmin0.x + ixw*binsize;
            bincorner.y = rmin0.y + iyw*binsize;
            bincorner.z = rmin0.z + iz*binsize;
            BinRoot[binnum] = 
				BuildTree(ParticleBin[binnum],theta,bincorner,binsize);

        }
    }
    }
    }
    TIMER_STOP(1);

	bcount = 0;
	for(iz=iz0; iz<iz1; iz++) {
	for(iy=iy0; iy<iy1; iy++) {
	for(ix=ix0; ix<ix1; ix++) {

/* Establish particle groups */
		binnum = ix + nx*(iy + ny*iz);
		if( BinRoot[binnum] != NULL ) {
			GroupParticles(BinRoot[binnum]);
		}
		else {
			SingleGroup(ParticleBin[binnum]);
		}
		ZeroGroups();

#ifdef DEBUG
		if( bcount % 1000 == 0 ) 
			fprintf(stderr,"P%d bin %d %d %d processing %d groups\n",
				myid,ixw,iyw,iz,NGroup);
#endif
	
		icount = 0;
		for(k=iz-1; k<=iz+1; k++) {
		for(j=iy-1; j<=iy+1; j++) { wrap(j0,j,ny);
		for(i=ix-1; i<=ix+1; i++) { wrap(i0,i,nx);
			if( k<0 || k>=nz ) continue;
			binnum = i0+nx*(j0 + ny*k);
			if( BinRoot[binnum] == NULL ) {
				if( ParticleBin[binnum] != NULL ) {
					GroupDirectForce(ParticleBin[binnum],prad2);
				}
			}
			else {
				GroupForce(BinRoot[binnum],thetasq);
			}
		}
		}
		}
		bcount++;
	}
	}
	}
#ifdef DEBUG
	fprintf(stderr,"P%d exiting force loop\n",myid);
#endif

	j = 0;
/*
	for(i=0, p=theParticles; p<theParticles+NParticles; p++,i++) {
*/
	for(iz=iz0; iz<iz1; iz++) {
	for(iy=iy0; iy<iy1; iy++) {
	for(ix=ix0; ix<ix1; ix++) {
		binnum = ix + nx*(iy + ny*iz);
		for(p=ParticleBin[binnum]; p != NULL; p=p->nextinbin ) {
			if( p->visitor ) continue;
			bp = p->bp;
#ifndef VarPM
			bp->freework = p->mono + 3*(p->quad);
#endif
			monotot += p->mono;
			quadtot += p->quad;
			if( p->mono < 0 ) {
				fprintf(stderr,"p->a.x %g p->a.y %g p->a.z %g\n",
					p->a.x, p->a.y, p->a.z);
			}
			bp->ax += (p->a.x)/L;
			bp->ay += (p->a.y)/L;
			bp->az += (p->a.z)/L;
#ifndef OBSERVATION_MODE
			/* half step evolution of Velocity */
			/* you must edit these lines for half step integration */
			bp->vx = fact1*(bp->vx) + fact2*(bp->ax);
			bp->vy = fact1*(bp->vy) + fact2*(bp->ay);
			bp->vz = fact1*(bp->vz) + fact2*(bp->az);
#endif
			j++;
		}
	}
	}
	}
	checkcount += j;

/* INSERT SPH CALL */

#ifdef SPH
	SphForces();
#endif


	FreeTreeMemory();

	}
	}
/* Rescale x,y,z's to PM scaling */
#ifdef DEBUG
	fprintf(stderr,"checkcount %d NBasicParticles %d NVisitors %d\n",checkcount,
		NBasicParticles,NVisitors);
#endif
#ifdef ACCOUT
	sprintf(accname,"acc.%.3d",myid);
	accfile = fopen(accname,"w");
	for(bp=BasicParticles; bp<BasicParticles+(size_t)NBasicParticles; bp++) {
			fprintf(accfile,"%d %g %g %g %g %g %g\n",
				bp->indx, (bp->x), bp->y, bp->z,(bp->ax)*L,(bp->ay)*L,(bp->az)*L);
	}
	fclose(accfile);
#endif

#ifndef VarPM
	totalwork = 0;
#endif
	disp = tdisp = 0.;
	for(i=0, bp=BasicParticles; bp<BasicParticles+(size_t)NBasicParticles; bp++,i++) {
#ifndef VarPM
		totalwork += bp->freework;
#endif
		bp->x  = ((bp->x) + 0.5)*L;
		bp->y  = ((bp->y) + 0.5)*L;
		bp->z  = ((bp->z) + 0.5)*L;

#ifndef OBSERVATION_MODE
		/* you must edit these lines for half step integration */
#ifdef GINIT
        bp->x -= ((bp->vx)*pfact);
        bp->y -= ((bp->vy)*pfact);
        bp->z -= ((bp->vz)*pfact);
#else
        bp->x += ((bp->vx)*pfact);
        bp->y += ((bp->vy)*pfact);
        bp->z += ((bp->vz)*pfact);
#endif
#endif
	disp = sqrt(pow(bp->vx*pfact,2)+pow(bp->vy*pfact,2)+pow(bp->vz*pfact,2));
	tdisp = MAX(tdisp,disp);

	}
	MPI_Allreduce(&tdisp,&disp,1,MPI_DOUBLE,MPI_MAX,MPI_COMM_WORLD);
	if(myid == 0) fprintf(stdout,"Maximum displacement %g %g\n",a,disp);
	TIMER_STOP(2);
#ifndef VarPM
	procspeed = totalwork/ELAPSED_TIME(2);
#endif
#ifdef DEBUG
	fprintf(stderr,"P%d Elapsed Time for Tree walks %g speed %g\n",
		myid, ELAPSED_TIME(2),procspeed);
#endif
#ifndef DYNAMIC_LOAD_BALANCE
	procspeed = 1.0;
#endif
#ifndef VarPM
	for(bp=BasicParticles; bp<BasicParticles+(size_t)NBasicParticles; bp++) {
			bp->freework /= procspeed;
	}
#endif

#ifdef DEBUG
	fprintf(stderr,"P%d Interaction list %g interactions/particle %g\n",
			myid, monotot+quadtot, (monotot+quadtot)/NBasicParticles);
#endif
	fprintf(stderr,"P%d Interaction list mono=%g quad=%g interactions/particle=%g\n", myid, monotot, quadtot, (monotot+quadtot)/NBasicParticles);

	SYNCHRONIZE();
TIMER_STOP(55);
if( myid == 0 )
	fprintf(stderr,"CPU(Tree) %g\n",ELAPSED_TIME(55));

	treestepcount++;
/*
	*nobj = NBasicParticles;
*/

#ifdef ACCOUT
	MPI_Finalize();
	exit(0);
#endif
	particles = (basicparticletype *)
		resizelast(particles,(size_t)NBasicParticles*sizeof(basicparticletype));
	{
		float axmin,aymin,azmin;
		float axmax,aymax,azmax;
		axmin = aymin = azmin = 1.e23;
		axmax = aymax = azmax = -1.e23;
		for(i=0;i<NBasicParticles;i++){
			axmin = MIN(axmin,particles[i].ax);
			aymin = MIN(aymin,particles[i].ay);
			azmin = MIN(azmin,particles[i].az);
			axmax = MAX(axmax,particles[i].ax);
			aymax = MAX(aymax,particles[i].ay);
			azmax = MAX(azmax,particles[i].az);
		}
		printf("P%d has x=%g %g y=%g %g z=%g %g\n",myid,axmin,axmax,
				aymin,aymax,azmin,azmax);
	}

	return NBasicParticles;
}

RescaleParticles(NBasicParticles,BasicParticles,scalingflag)
int NBasicParticles;
basicparticletype *BasicParticles;
int scalingflag;
{
	basicparticletype *p;

	if( scalingflag == TO_TREE_SCALING ) {
/* Scale so particles are in a box from -0.5 to 0.5 */
		for(p=BasicParticles; p<BasicParticles+(size_t)NBasicParticles; p++) {
			p->x = (p->x)/L - 0.5;
			p->y = (p->y)/L - 0.5;
			p->z = (p->z)/L - 0.5;
			/*
            if( p->x > 0.5 ) { p->x -= 1.0; } 
				else if( p->x < -0.5 ) { p->x += 1.0; };
            if( p->y > 0.5 ) { p->y -= 1.0; } 
				else if( p->y < -0.5 ) { p->y += 1.0; };
            if( p->z > 0.5 ) { p->z -= 1.0; } 
				else if( p->z < -0.5 ) { p->z += 1.0; };
				*/
            if( p->x >= xboxup ) { p->x -= xboxsize; } 
				else if( p->x < -0.5 ) { p->x += xboxsize; };
            if( p->y >= yboxup ) { p->y -= yboxsize; } 
				else if( p->y < -0.5 ) { p->y += yboxsize; };
            if( p->z >= zboxup ) { p->z -= zboxsize; } 
				else if( p->z < -0.5 ) { p->z += zboxsize; };

		}
	}
	else if( scalingflag == TO_PM_SCALING ) {
		for(p=BasicParticles; p<BasicParticles+(size_t)NBasicParticles; p++) {
			p->x = ((p->x) + 0.5)*L;
			p->y = ((p->y) + 0.5)*L;
			p->z = ((p->z) + 0.5)*L;
		}
	}
}

InitializeTreeParticles(BasicParticles,ix0,ix1,iy0,iy1,iz0,iz1)
basicparticletype *BasicParticles;
int ix0, ix1, iy0, iy1, iz0, iz1;
{
	int ix, iy, iz, ixw, iyw;
	int count, tstcount;
	int binnum;
	basicparticletype *bp;
	particleStruct *p, *p0;

	count = 0;
	for(iz=iz0; iz<iz1; iz++) {
	for(iy=iy0; iy<iy1; iy++) { wrap(iyw,iy,ny);
	for(ix=ix0; ix<ix1; ix++) { wrap(ixw,ix,nx);
		binnum = ixw + nx*(iyw + ny*iz);
if( binnum > NBins ) fprintf( stderr,"binnum=%d > NBins=%d\n",binnum,NBins);
		count += BinCount[binnum];
	}
	}
	}

	NParticles = count;
	theParticles = (particleStruct *) Malloc(count*sizeof(particleStruct),
		PPTR(theParticles));
	InitCellMemory(2*count);

	p = theParticles;
	for(iz=iz0; iz<iz1; iz++) {
	for(iy=iy0; iy<iy1; iy++) { wrap(iyw,iy,ny);
	for(ix=ix0; ix<ix1; ix++) { wrap(ixw,ix,nx);
		binnum = ixw + nx*(iyw + ny*iz);
		ParticleBin[binnum] = NULL;
		BinRoot[binnum] = NULL;
		if( BasicParticleBin[binnum] != NULL ) {
			ParticleBin[binnum] = p;
			tstcount = 0;
			for(bp=BasicParticleBin[binnum]; bp != NULL; bp=bp->u.nextinbin ) {
				p->type = PARTICULE_TYPE;
/*
#ifdef COMPACT 
				p->mass = pmass0;
				p->epssq = epssq0;
#else
				p->mass = (bp->mass);
				p->epssq = (bp->epssq);
#endif
*/
#if defined(MULTIMASS)
				p->mass = (bp->mass);
#else
				p->mass = pmass0;
#endif
#if defined(INDVEPS)
				p->epssq = (bp->epssq);
#else
				p->epssq = epssq0;
#endif


				p->quad = p->mono = 0;
				p->a.x = p->a.y = p->a.z = 0;
				p->r.x = (bp->x);
				p->r.y = (bp->y);
				p->r.z = (bp->z);
#ifndef NOPINDEX
				p->idnum = (bp->indx);
#endif
#ifndef VarPM
				p->work = (bp->freework);
#endif
				if( bp < BasicParticles+(size_t)NBasicParticles )
					p->visitor = 0;
				else
					p->visitor = 1;
#ifdef never
				if( (ix==ix0) || (ix==(ix1-1)) ) p->visitor = 1;
				if( (iy==iy0) || (iy==(iy1-1)) ) p->visitor = 1;
#endif
				p->bp = bp;
				p->nextinbin = p+1;
				p++;
				tstcount++;
			}
			if( tstcount != BinCount[binnum] )
				fprintf(stderr,"particle bin misset tstcount %d BinCount %d\n",
					tstcount, BinCount[binnum]);
			p--; 
			p->nextinbin = NULL;
			p++;
			count = 0;
			for(p0=ParticleBin[binnum]; p0 != NULL; p0=p0->nextinbin )
				count++;
			if( count != BinCount[binnum] ) {
				fprintf(stderr,"particle bin misset count %d BinCount %d\n",
					count, BinCount[binnum]);
			}
		}
	}
	}
	}
}

