     !Simple test program to print out sigma_8 as a function of the CDM density
      module MyModule
         use CAMB
         type (CAMBdata) OutData,InData
      end module MyModule

      program InitPower
        use CAMB
        use MyModule
        use InitialPower
        implicit none
        integer i,j,k,nmatterpk
        real PI2,rng,maxkhfactor
        parameter(PI2=3.1415926535d0*2.d0,maxkhfactor=5)
    
        real matterpk(8192),rkarray(8192)
        type(CAMBparams) P !defined in ModelParams in modules.f90
        !type (CAMBdata) OutData
        type(MatterTransferData) MTrans
        real omep,omepb,omeplam,hubble,npower,maxkh,redshift,boxsize
        integer itf,in
        integer error
        real minkh,dlnkh,pamp,bias,khmax,opamp
        integer npoints
        character(LEN=80) fmt,inputfile,outfile
        logical OK
        double precision cubicbox



        call getarg(1,inputfile)
       open(1,file=inputfile,form='unformatted')
!        read(1) pamp
        read(1) InData%Params
        CP = Indata%Params
        read(1) InData%MTrans%num_q_trans
        call Transfer_Allocate(InData%MTrans)
        read(1) (InData%MTrans%q_trans(i),i=1,InData%MTrans%num_q_trans)
        read(1) (((InData%MTrans%TransferData(i,j,k),&
                i=1,Transfer_max), &
                j=1,InData%MTrans%num_q_trans), &
                k=1,CP%Transfer%num_redshifts)
        read(1) ((InData%MTrans%sigma_8(i,j), &
                i=1,CP%Transfer%num_redshifts), &
                j=1,CP%InitPower%nn)
        close(1)
		pamp = InData%MTrans%sigma_8(1,1)**2

!        print *,CP%Transfer%redshifts(1),Indata%Params%Transfer%redshifts(1)
!        boxsize = 128
!        print *,pamp, InData%MTrans%sigma_8(1,1)**2*boxsize**3

       
        omepb = CP%omegab 
        omep = CP%omegac  + omepb
        omeplam  = CP%omegav  
        hubble = CP%H0  /100.
        write(*,*)
        write(*,*)
        print *, '#####################################################'
        print *, '####      POWER SPECTRUM PARAMETER        ###########'
        print *, '#####################################################'
        print *, 'omep                  = ',omep
        print *, 'omepb                 = ',omepb
        print *, 'omeplam               = ',omeplam
        print *, 'npower                = ',CP%InitPower%an(1)
        print *, 'Hubble (km/sec.)      = ',CP%H0
        if(CP%OutputNormalization) then
        print *, 'Normalization applied   '
        else 
        print *, 'Normalization unapplied '
        endif
        print *, 'simulation amplitude/boxsize**3  = ',pamp
        print *, 'power Amplitude       = ',CP%InitPower%ScalarPowerAmp(1)
        print *, 'number of redshift    = ',CP%InitPower%nn
        print *, 'Redshift              = ',CP%Transfer%redshifts(1)
        print *, 'Highprecision         = ',CP%Transfer%high_precision
        print *, 'kmax                  = ',CP%Transfer%kmax
        print *, 'k per logint          = ',CP%Transfer%k_per_logint
        print *, '#####################################################'
        print *, '####      POWER SPECTRUM PARAMETER        ###########'
        print *, '#####################################################'
        write(*,*)
        write(*,*)

        OK= CAMB_ValidateParams(CP)
        if(OK .eq. .false.) then
           print *,'###################################'
           print *,'###################################'
           print *,'Wrong set of parameters in the CAMB'
           print *,'###################################'
           print *,'###################################'
           print *,'Ok = ',OK
           stop
        endif

        P = CP
        call InitializePowers(CP%InitPower,CP%curv)

         nmatterpk = 8192
         call GetPower(minkh,dlnkh,matterpk,nmatterpk)
         do i = 1, nmatterpk
            rkarray(i) = minkh*exp((i-1)*dlnkh)
!            matterpk(i) = sqrt(matterpk(i))
            write(*,*) rkarray(i), matterpk(i)
         enddo


        !call CAMB_GetResults(P) 
!        call CAMB_GetTransfers(P,OutData)
             
!       This cubic box multiplication is to normalize the amplitude in the unit of 2pi*u not k.
!        pamp = (MT%sigma_8(1,1))**2*cubicbox

!       open(1,file='Check.dat',form='unformatted')
!       read(1) P,opamp
!        read(1) InData%Params
!        read(1) InData%MTrans%num_q_trans
!        call Transfer_Allocate(InData%MTrans)
!        read(1) (InData%MTrans%q_trans(i),i=1,InData%MTrans%num_q_trans)
!        read(1) (((InData%MTrans%TransferData(i,j,k),i=1,Transfer_max),j=1,InData%MTrans%num_q_trans), &
!                k=1,CP%Transfer%num_redshifts)
!        read(1) ((InData%MTrans%sigma_8(i,j),i=1,CP%Transfer%num_redshifts),j=1,CP%InitPower%nn)
!        close(1)


       
      end program InitPower

      subroutine GetPower(minkh,dlnkh,matterpk,nmatterpk)
        use CAMB
        use MyModule
        use transfer
!        use InitialPower
        implicit none
        integer i
    
        type(MatterTransferData) MTrans
        integer itf,in
        integer error
        real minkh,dlnkh
        integer npoints,nmatterpk
        real, dimension(:,:), allocatable :: outpower
        real matterpk(nmatterpk)
        character(LEN=80) fmt
        logical OK


        minkh = 5e-5
        dlnkh = 0.02
        Mtrans = InData%MTrans


        itf = 1

!        print *, Mtrans%num_q_trans,MTrans%TransferData(Transfer_kh,Mtrans%num_q_trans,itf)

         npoints = alog(MTrans%TransferData(Transfer_kh,Mtrans%num_q_trans,itf)/minkh)/dlnkh+1
         if(npoints .gt. nmatterpk) then
            print *,'#################################################'
            print *,'#################################################'
            print *,'Small number of power spectrum array ',nmatterpk
            print *,'The npoints in CAMB is ', npoints
            print *,'#################################################'
            print *,'#################################################'
            stop
         else
            nmatterpk = npoints
         endif
         in = 1
         call Transfer_GetMatterPower(Mtrans,matterpk(1),itf,in,minkh,dlnkh,npoints)

        end subroutine GetPower

